"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import json
import os
import re
import tarfile
import tempfile

import docker
import docker.errors

from buildrunner.docker import new_client, force_remove_container


class DockerBuilder:  # pylint: disable=too-many-instance-attributes
    """
    An object that manages and orchestrates building a Docker image from
    a Dockerfile.
    """

    def __init__(
            self,
            path=None,
            inject=None,
            dockerfile=None,
            dockerd_url=None,
            timeout=None,
            docker_registry=None,
            temp_dir=None,
    ):  # pylint: disable=too-many-arguments
        self.path = path
        self.inject = inject
        self.temp_dir = temp_dir
        self.dockerfile = None
        self.cleanup_dockerfile = False
        if dockerfile:
            if os.path.exists(dockerfile):
                self.dockerfile = dockerfile
            else:
                df_file = tempfile.NamedTemporaryFile(delete=False, dir=self.temp_dir)
                try:
                    df_file.write(dockerfile.encode('utf-8'))
                    self.cleanup_dockerfile = True
                    self.dockerfile = df_file.name
                finally:
                    df_file.close()

        self.docker_client = new_client(
            dockerd_url=dockerd_url,
            timeout=timeout,
        )
        self.docker_registry = docker_registry
        self.image = None
        self.intermediate_containers = []

    @staticmethod
    def _sanitize_buildargs(buildargs=None):
        """
        Ensure that buildargs are correct for the Docker API.

        Args:

            :buildargs: (dict): unsanitized arguments to be passed to the Docker client.

        Returns:

            :dict: sanitized arguments to be passed to the Docker client.
        """
        if not isinstance(buildargs, dict):
            raise TypeError('buildargs must be a dictionary of keys/values')

        return {k: str(v) for k, v in list(buildargs.items())}

    # pylint: disable=too-many-branches,too-many-locals,too-many-arguments,invalid-name
    def build(self, console=None, nocache=False, cache_from=None, rm=True, pull=True, buildargs=None, platform=None):
        """
        Run a docker build using the configured context, constructing the
        context tar file if necessary.
        """
        if cache_from is None:
            cache_from = []
        if buildargs is None:
            buildargs = {}
        # create our own tar file, injecting the appropriate paths
        _fileobj = tempfile.NamedTemporaryFile(dir=self.temp_dir)
        tfile = tarfile.open(mode='w', fileobj=_fileobj)
        if self.path:
            tfile.add(self.path, arcname='.')
        if self.inject:
            for to_inject, dest in self.inject.items():
                tfile.add(to_inject, arcname=dest)
        if self.dockerfile:
            tfile.add(self.dockerfile, arcname='./Dockerfile')
        tfile.close()
        _fileobj.seek(0)

        # Always add default registry to build args
        buildargs['DOCKER_REGISTRY'] = self.docker_registry

        stream = self.docker_client.build(
            path=None,
            nocache=nocache,
            cache_from=cache_from,
            custom_context=True,
            fileobj=_fileobj,
            rm=rm,
            pull=pull,
            buildargs=self._sanitize_buildargs(buildargs),
            platform=platform
        )

        # monitor output for logs and status
        exit_code = 0
        msg_buffer = ''
        for msg_str in stream:  # pylint: disable=too-many-nested-blocks
            for msg in msg_str.decode('utf-8').split("\n"):
                if msg:
                    msg_buffer += msg
                    try:
                        # there is a limit on the chars returned in the stream
                        # generator, so if we don't have a valid json message
                        # here we get the next msg and append to the current
                        # one
                        json_msg = json.loads(msg_buffer)
                        msg_buffer = ''
                    except ValueError:
                        continue
                    if 'stream' in json_msg:
                        # capture intermediate containers for cleanup later
                        # the command line 'docker build' has a '--force-rm' option,
                        # but that isn't available in the python client
                        container_match = re.search(
                            r' ---> Running in ([0-9a-f]+)',
                            json_msg['stream'],
                        )
                        if container_match:
                            self.intermediate_containers.append(
                                container_match.group(1)
                            )

                        # capture the resulting image
                        image_match = re.search(
                            r'Successfully built ([0-9a-f]+)',
                            json_msg['stream'],
                        )
                        if image_match:
                            self.image = image_match.group(1)

                        if console:
                            console.write(json_msg['stream'])
                    if 'error' in json_msg:
                        exit_code = 1
                        if 'errorDetail' in json_msg:
                            if 'message' in json_msg['errorDetail'] and console:
                                console.write(json_msg['errorDetail']['message'])
                                console.write('\n')

        return exit_code

    def cleanup(self):
        """
        Cleanup the docker build environment.
        """
        # cleanup the generated dockerfile if present
        if self.cleanup_dockerfile:
            if self.dockerfile and os.path.exists(self.dockerfile):
                os.remove(self.dockerfile)

        # iterate through and destory intermediate containers
        for container in self.intermediate_containers:
            try:
                force_remove_container(self.docker_client, container)
            except docker.errors.APIError:
                pass



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""
# pylint: disable=too-many-lines

import base64
import codecs
from collections import OrderedDict
import copy
import datetime
import errno
import fnmatch
import getpass
from graphlib import TopologicalSorter
import importlib.machinery
import inspect
from io import StringIO
import json
import logging
import os
import re
import shutil
import sys
import tarfile
import tempfile
import types

import jinja2
import requests

from vcsinfo import detect_vcs, VCSUnsupported, VCSMissingRevision

from buildrunner import docker
from buildrunner.docker.builder import DockerBuilder
from buildrunner.errors import (
    BuildRunnerConfigurationError,
    BuildRunnerProcessingError,
    BuildRunnerVersionError,
    ConfigVersionFormatError,
    ConfigVersionTypeError,
)
from buildrunner.sshagent import load_ssh_key_from_file, load_ssh_key_from_str
from buildrunner.steprunner import BuildStepRunner
from buildrunner.steprunner.tasks.push import sanitize_tag
from buildrunner.utils import (
    ConsoleLogger,
    checksum,
    epoch_time,
    hash_sha1,
    load_config,
)
from . import fetch
# from . import BuildRunnerVersionError, ConfigVersionFormatError, ConfigVersionTypeError

LOGGER = logging.getLogger(__name__)

__version__ = 'DEVELOPMENT'
try:
    _VERSION_FILE = os.path.join(os.path.dirname(__file__), 'version.py')
    if os.path.exists(_VERSION_FILE):
        loader = importlib.machinery.SourceFileLoader('buildrunnerversion', _VERSION_FILE)
        _VERSION_MOD = types.ModuleType(loader.name)
        loader.exec_module(_VERSION_MOD)
        __version__ = getattr(_VERSION_MOD, '__version__', __version__)
except:
    pass

MASTER_GLOBAL_CONFIG_FILE = '/etc/buildrunner/buildrunner.yaml'
DEFAULT_GLOBAL_CONFIG_FILES = [
    MASTER_GLOBAL_CONFIG_FILE,
    '~/.buildrunner.yaml',
]

DEFAULT_CACHES_ROOT = '~/.buildrunner/caches'
DEFAULT_RUN_CONFIG_FILES = ['buildrunner.yaml', 'gauntlet.yaml']
RESULTS_DIR = 'buildrunner.results'

SOURCE_DOCKERFILE = os.path.join(os.path.dirname(__file__), 'SourceDockerfile')


class BuildRunner:  # pylint: disable=too-many-instance-attributes
    """
    Class used to manage running a build.
    """

    CONTEXT_ENV_PREFIXES = ['BUILDRUNNER_', 'VCSINFO_', 'PACKAGER_', 'GAUNTLET_']

    def _get_config_context(self, ctx=None):
        """
        Generate the Jinja configuration context for substitution

        Args:
          ctx (dict): A dictionary of key/values to be merged
          over the default, generated context.

        Returns:
          dict: A dictionary of key/values to be substituted
        """

        context = {
            'BUILDRUNNER_BUILD_NUMBER': str(self.build_number),
            'BUILDRUNNER_BUILD_ID': str(self.build_id),
            'BUILDRUNNER_BUILD_DOCKER_TAG': str(sanitize_tag(self.build_id)),
            'BUILDRUNNER_BUILD_TIME': str(self.build_time),
            'VCSINFO_NAME': str(self.vcs.name),
            'VCSINFO_BRANCH': str(self.vcs.branch),
            'VCSINFO_NUMBER': str(self.vcs.number),
            'VCSINFO_ID': str(self.vcs.id),
            'VCSINFO_SHORT_ID': str(self.vcs.id)[:7],
            'VCSINFO_MODIFIED': str(self.vcs.modified),
            'BUILDRUNNER_STEPS': self.steps_to_run,
        }

        if ctx:
            context.update(ctx)

        for env_name, env_value in os.environ.items():
            for prefix in self.CONTEXT_ENV_PREFIXES:
                if env_name.startswith(prefix):
                    context[env_name] = env_value

        return context

    @staticmethod
    def _raise_exception_jinja(message):
        """
        Raises an exception from a jinja template.
        """
        raise Exception(message)

    def _log_generated_file(self, file_name, file_contents):
        '''
        Conditionally log the contents of a generated file.
        '''
        if self.log_generated_files:
            self.log.write(f'Generated contents of {file_name}:\n')
            for line in file_contents.splitlines():
                self.log.write(f'{line}\n')

    def _read_yaml_file(self, filename):
        """
        Reads a file in the local workspace as Jinja-templated
        YAML and returns the contents.
        Throws an error on failure.
        """
        with codecs.open(filename, 'r', encoding='utf-8') as _file:
            jtemplate = jinja2.Template(_file.read())
        context = copy.deepcopy(self.env)
        file_contents = jtemplate.render(context)
        self._log_generated_file(filename, file_contents)
        return load_config(StringIO(file_contents), filename)

    def _load_config_files(self, cfg_files=None, ctx=None, log_file=True):
        """
        Load config files templating them with Jinja and parsing the YAML.

        Args:
            cfg_files (list):  List of configuration files
            ctx (dict): context object to load into config_context
            log_file (boolean): will write out the generated config to log(s)
                Default: True

        Returns:
          multi-structure: configuration keys and values
        """

        cfg_files = cfg_files or []

        username = getpass.getuser()
        homedir = os.path.expanduser('~')

        context = ctx or {}
        for cfg in cfg_files:
            cfg_path = os.path.realpath(os.path.expanduser(cfg))
            if os.path.exists(cfg_path):
                ctx = self._load_config(cfg_path, context, log_file=log_file)
                if ctx is None:
                    # Empty config file
                    continue

                # Only allow MASTER_GLOBAL_CONFIG_FILE to specify arbitrary local-files for mounting
                # - all other local-files get scrubbed for specific requirements and non-matches
                # are dropped.
                if cfg_path != MASTER_GLOBAL_CONFIG_FILE:
                    scrubbed_local_files = {}
                    for fname, fpath in list(ctx.get('local-files', {}).items()):
                        if not isinstance(fpath, str):
                            self.log.write(
                                f'Bad "local-files" entry in {cfg_path!r}:\n'
                                f'    {fname!r}: {fpath!r}\n'
                            )
                            continue
                        resolved_path = os.path.realpath(os.path.expanduser(fpath))
                        # pylint: disable=too-many-boolean-expressions
                        if (
                                username == 'root'
                                or resolved_path == homedir
                                or resolved_path.startswith(homedir + os.path.sep)
                                or os.stat(resolved_path).st_uid == os.getuid()
                                or (
                                    not os.path.isdir(resolved_path)
                                    and os.access(resolved_path, os.R_OK | os.W_OK)
                                )
                        ):
                            scrubbed_local_files[fname] = resolved_path
                        else:
                            self.log.write(
                                f'Bad "local-files" entry in {cfg_path!r}:\n'
                                f'    User {username!r} is not allowed to mount {resolved_path!r}.\n'
                                f'    You may need an entry in {MASTER_GLOBAL_CONFIG_FILE!r}.\n'
                            )
                    ctx['local-files'] = scrubbed_local_files

                context.update(ctx)

        context.update({'CONFIG_FILES': cfg_files})

        return context

    def _strftime(self, _format="%Y-%m-%d", _ts=None):
        """
        Format the provided timestamp. If no timestamp is provided, build_time is used
        :param _format: Format string - default "%Y-%m-%d"
        :param _ts: Timestamp to format - default self.build_time
        :return: Formatted date/time string
        """
        if _ts is None:
            _ts = self.build_time
        _date = datetime.datetime.fromtimestamp(_ts)
        return _date.strftime(_format)

    @staticmethod
    def _re_sub_filter(text, pattern, replace, count=0, flags=0):
        '''
        Filter for regular expression replacement.
        :param text: The string being examined for ``pattern``
        :param pattern: The pattern to find in ``text``
        :param replace: The replacement for ``pattern``
        :param count: How many matches of ``pattern`` to replace with ``replace`` (0=all)
        :param flags: Regular expression flags
        '''
        return re.sub(pattern, replace, text, count=count, flags=flags)

    @staticmethod
    def _re_split_filter(text, pattern, maxsplit=0, flags=0):
        '''
        Filter for regular expression replacement.
        :param text: The string being examined for ``pattern``
        :param pattern: The pattern used to split ``text``
        :param maxsplit: How many instances of ``pattern`` to split (0=all)
        :param flags: Regular expression flags
        '''
        return re.split(pattern, text, maxsplit=maxsplit, flags=flags)

    @staticmethod
    def _reorder_dependency_steps(config):
        """
        Reorders the steps based on the dependencies that are outlined in the config
        """
        # Defines configuration keywords, should add to a config validation class
        keyword_version = 'version'
        keyword_steps = 'steps'
        keyword_depends =   'depends'
        supported_version = 2.0

        if keyword_version not in config.keys() \
                or config[keyword_version] < supported_version:
            return config

        ordered_steps = OrderedDict()
        if keyword_steps in config.keys():
            topo_sorter = TopologicalSorter()
            for name, instructions in config[keyword_steps].items():
                if keyword_depends in instructions.keys():
                    for depend in instructions[keyword_depends]:
                        topo_sorter.add(name, depend)
                else:
                    topo_sorter.add(name)
            for step in topo_sorter.static_order():
                if step not in config[keyword_steps]:
                    raise KeyError(f"Step '{step}' is not defined and is listed as a step dependency in "
                                   f"the configuration. "
                                   f"Please correct the typo or define step '{step}' in the configuration.")

                if keyword_depends in config[keyword_steps][step].keys():
                    del config[keyword_steps][step][keyword_depends]

                ordered_steps[step] = config[keyword_steps][step]

            config[keyword_steps] = ordered_steps

        return config


    @staticmethod
    def _validate_version(config: OrderedDict, version_file_path: str):
        """
        Compares that the version in the config is less than or equal to the current version of
        buildrunner. If the config version is greater than the buildrunner version or any parsing error occurs
        it will raise a buildrunner exception.
        """
        buildrunner_version = None

        if not os.path.exists(version_file_path):
            print(f"WARNING: File {version_file_path} does not exist. This could indicate an error with "
                  f"the buildrunner installation. Unable to validate version.")
            return

        with open(version_file_path, 'r') as version_file:
            for line in version_file.readlines():
                if '__version__' in line:
                    try:
                        version_values = line.split('=')[1].strip().replace("'","").split('.')
                        buildrunner_version = f"{version_values[0]}.{version_values[1]}"
                    except IndexError as exception:
                        raise ConfigVersionFormatError(f"couldn't parse version from \"{line}\"") from exception

        if not buildrunner_version:
            raise BuildRunnerVersionError("unable to determine buildrunner version")

        config_version = None

        # version is optional and is valid to not have it in the config
        if 'version' not in config.keys():
            return

        config_version = config['version']

        try:
            if float(config_version) > float(buildrunner_version):
                raise ConfigVersionFormatError(f"configuration version {config_version} is higher than "
                                         f"buildrunner version {buildrunner_version}")
        except ValueError as exception:
            raise ConfigVersionTypeError(f"unable to convert config version \"{config_version}\" "
                                         f"or buildrunner version \"{buildrunner_version}\" "
                                         f"to a float") from exception


    def _load_config(self, cfg_file, ctx=None, log_file=True):
        """
        Load a config file templating it with Jinja and parsing the YAML.

        Returns:
          multi-structure: configuration keys and values
        """

        config = None
        fetch_file = cfg_file
        visited = set()

        while True:
            visited.add(fetch_file)
            contents = fetch.fetch_file(fetch_file, self.global_config)
            jenv = jinja2.Environment(loader=jinja2.FileSystemLoader('.'), extensions=['jinja2.ext.do'])
            jenv.filters['hash_sha1'] = hash_sha1
            jenv.filters['base64encode'] = base64.encode
            jenv.filters['base64decode'] = base64.decode
            jenv.filters['re_sub'] = self._re_sub_filter
            jenv.filters['re_split'] = self._re_split_filter

            jenv.globals.update(checksum=checksum)
            jtemplate = jenv.from_string(contents)

            config_context = copy.deepcopy(self.env)
            config_context.update({
                'CONFIG_FILE': cfg_file,
                'CONFIG_DIR': os.path.dirname(cfg_file),
                # This is stored after the initial env is set
                'DOCKER_REGISTRY': self.get_docker_registry(),
                'read_yaml_file': self._read_yaml_file,
                'raise': self._raise_exception_jinja,
                'strftime': self._strftime,
            })

            if ctx:
                config_context.update(ctx)

            config_contents = jtemplate.render(config_context)
            if log_file:
                self._log_generated_file(cfg_file, config_contents)
            config = load_config(StringIO(config_contents), cfg_file)

            if not config:
                break

            redirect = config.get('redirect')
            if redirect is None:
                break

            fetch_file = redirect
            if fetch_file in visited:
                raise BuildRunnerConfigurationError(
                    f"Redirect loop visiting previously visited file: {fetch_file}"
                )

        self._validate_version(config=config,
                               version_file_path=f"{os.path.dirname(os.path.realpath(__file__))}/version.py")

        config = self._reorder_dependency_steps(config)

        return config

    def __init__(
            self,
            build_dir,
            build_results_dir=None,
            global_config_file=None,
            run_config_file=None,
            run_config=None,
            build_number=None,
            push=False,
            colorize_log=True,
            cleanup_images=False,
            cleanup_step_artifacts=False,
            steps_to_run=None,
            publish_ports=False,
            disable_timestamps=False,
            log_generated_files=False,
            docker_timeout=None,
            local_images=False,
            platform=None,
    ):  # pylint: disable=too-many-statements,too-many-branches,too-many-locals,too-many-arguments
        """
        """
        self.build_dir = build_dir
        if build_results_dir:
            self.build_results_dir = build_results_dir
        else:
            self.build_results_dir = os.path.join(self.build_dir, RESULTS_DIR)
        self.push = push
        self.cleanup_images = cleanup_images
        self.cleanup_step_artifacts = cleanup_step_artifacts
        self.generated_images = []
        # The set of images (including tag) that were committed as part of this build
        # This is used to check if images should be pulled by default or not
        self.committed_images = set()
        self.repo_tags_to_push = []
        self.colorize_log = colorize_log
        self.steps_to_run = steps_to_run
        self.publish_ports = publish_ports
        self.disable_timestamps = disable_timestamps
        self.log_generated_files = log_generated_files
        self.docker_timeout = docker_timeout
        self.local_images = local_images
        self.platform = platform

        self.tmp_files = []
        self.artifacts = OrderedDict()
        self.pypi_packages = OrderedDict()

        self.exit_code = None
        self._source_image = None
        self._source_archive = None
        self._log_file = None
        self._log = None

        # set build time
        self.build_time = epoch_time()

        # set build number
        self.build_number = build_number
        if not self.build_number:
            self.build_number = self.build_time

        try:
            self.vcs = detect_vcs(self.build_dir)
            self.build_id = f"{self.vcs.id_string}-{self.build_number}"
        except VCSUnsupported as err:
            self.log.write(f'{err}\nPlease verify you have a VCS set up for this project.\n')
            sys.exit()
        except VCSMissingRevision as err:
            self.log.write(f'{err}\nMake sure you have at least one commit.\n')
            sys.exit()

        # cleanup existing results dir (if needed)
        if self.cleanup_step_artifacts and os.path.exists(self.build_results_dir):
            shutil.rmtree(self.build_results_dir)
            self.log.write(f'Cleaned existing results directory "{RESULTS_DIR}"')

        # default environment - must come *after* VCS detection
        base_context = {}
        if push:
            base_context['BUILDRUNNER_DO_PUSH'] = 1
        self.env = self._get_config_context(base_context)
        # pylint: disable=consider-iterating-dictionary
        key_len = max([len(key) for key in self.env.keys()])
        for key in sorted(self.env.keys()):
            val = self.env[key]
            LOGGER.debug(f'Environment: {key!s:>{key_len}}: {val}')

        # load global configuration
        _gc_files = DEFAULT_GLOBAL_CONFIG_FILES[:]
        _gc_files.append(global_config_file or f'{self.build_dir}/.buildrunner.yaml')

        _global_config_files = self.to_abs_path(
            _gc_files, return_list=True
        )

        self.log.write(f"\nGlobal configuration is from: {', '.join(_global_config_files)}\n")
        self.global_config = {}
        self.global_config = self._load_config_files(_global_config_files, log_file=False)

        # load run configuration
        _run_config_file = None
        if run_config:
            self.run_config = run_config
        else:
            if run_config_file:
                _run_config_file = self.to_abs_path(run_config_file)
            else:
                self.log.write("looking for run configuration\n")
                for name_to_try in DEFAULT_RUN_CONFIG_FILES:
                    _to_try = self.to_abs_path(name_to_try)
                    if os.path.exists(_to_try):
                        _run_config_file = _to_try
                        break

            if not _run_config_file or not os.path.exists(_run_config_file):
                raise BuildRunnerConfigurationError(
                    'Cannot find build configuration file'
                )

            self.run_config = self._load_config(_run_config_file)

        if not isinstance(self.run_config, dict) or 'steps' not in self.run_config:
            cfg_file = _run_config_file if _run_config_file else 'provided config'
            raise BuildRunnerConfigurationError(f'Could not find a "steps" attribute in {cfg_file}')
        if not self.run_config['steps'] or not isinstance(self.run_config['steps'], dict):
            cfg_file = _run_config_file if _run_config_file else 'provided config'
            raise BuildRunnerConfigurationError(
                f'The "steps" attribute is not a non-empty dictionary in {cfg_file}'
            )

    def get_docker_registry(self):
        """
        Default to docker.io if none is configured
        """
        return self.global_config.get('docker-registry', 'docker.io')

    def get_temp_dir(self):
        """
        Get temp dir in the following priorities:
        * Environment variable
        * Global configuration property
        * Configured system temp directory
        """
        return os.getenv('BUILDRUNNER_TEMPDIR', self.global_config.get('temp-dir', tempfile.gettempdir()))

    def get_build_server_from_alias(self, host):
        """
        Given a host alias string determine the actual build server host value
        to use by checking the global configuration.
        """
        # if no build servers configuration in global config just return host
        if 'build-servers' not in self.global_config:
            return host

        build_servers = self.global_config['build-servers']
        for _host, _host_aliases in build_servers.items():
            if host in _host_aliases:
                return _host

        return host

    def get_ssh_keys_from_aliases(self, key_aliases):  # pylint: disable=too-many-branches
        """
        Given a list of key aliases return Paramiko key objects based on keys
        registered in the global config.
        """
        ssh_keys = []
        if not key_aliases:
            return ssh_keys
        if 'ssh-keys' not in self.global_config:
            raise BuildRunnerConfigurationError(
                "SSH key aliases specified but no 'ssh-keys' "
                "configuration in global build runner config"
            )

        ssh_keys = self.global_config['ssh-keys']
        _keys = []
        _matched_aliases = []
        for key_info in ssh_keys:
            if 'aliases' not in key_info or not key_info['aliases']:
                continue

            _password = None
            _prompt_for_password = False
            if 'password' in key_info:
                _password = key_info['password']
            else:
                _prompt_for_password = key_info.get('prompt-password', False)

            for alias in key_aliases:
                if alias in key_info['aliases']:
                    _matched_aliases.append(alias)

                    # Prompt for password if necessary.  Only once per key
                    if _prompt_for_password:
                        _password = getpass.getpass(f"Password for SSH Key ({alias}): ")
                        _prompt_for_password = False

                    if 'file' in key_info:
                        _key_file = os.path.realpath(
                            os.path.expanduser(
                                os.path.expandvars(key_info['file'])
                            )
                        )

                        _keys.append(
                            load_ssh_key_from_file(_key_file, _password)
                        )
                    elif 'key' in key_info:
                        _keys.append(
                            load_ssh_key_from_str(key_info['key'], _password)
                        )

        for alias in key_aliases:
            if alias not in _matched_aliases:
                raise BuildRunnerConfigurationError(
                    f"Could not find valid SSH key matching alias '{alias}'"
                )

        return _keys

    def get_local_files_from_alias(self, file_alias):
        """
        Given a file alias lookup the local file path from the global config.
        """
        if not file_alias:
            return None
        if 'local-files' not in self.global_config:
            self.log.write("No 'local-files' configuration in global build runner config")
            return None

        local_files = self.global_config['local-files']
        for local_alias, local_file in local_files.items():
            if file_alias == local_alias:
                local_path = os.path.realpath(
                    os.path.expanduser(os.path.expandvars(local_file))
                )
                if os.path.exists(local_path):
                    return local_path

                # need to put the contents in a tmp file and return the path
                _fileobj = tempfile.NamedTemporaryFile(
                    delete=False,
                    dir=self.get_temp_dir(),
                )
                _fileobj.write(local_file)
                tmp_path = os.path.realpath(_fileobj.name)
                _fileobj.close()
                self.tmp_files.append(tmp_path)
                return tmp_path

        return None

    @staticmethod
    def get_cache_archive_ext():
        """
        Returns the archive file extension used for cache archive files
        """
        return "tar"

    def get_cache_archive_file(self, cache_name, project_name=""):
        """
        Given a cache name determine the local file path.
        """
        caches_root = self.global_config.get('caches-root', DEFAULT_CACHES_ROOT)
        build_path = os.path.splitdrive(self.build_dir)[1]
        if os.path.isabs(build_path):
            build_path = build_path[1:]

        cache_name = f"{cache_name}.{self.get_cache_archive_ext()}"
        if project_name != "":
            cache_name = f"{project_name}-{cache_name}"

        local_cache_archive_file = os.path.expanduser(
            os.path.join(caches_root, cache_name)
        )
        cache_dir = os.path.dirname(local_cache_archive_file)
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)
        return local_cache_archive_file

    def to_abs_path(self, path, return_list=False):
        """
        Convert a path to an absolute path (if it isn't one already).
        """

        paths = path
        if not isinstance(path, list):
            paths = [path]

        for index, _ in enumerate(paths):
            _path = os.path.expanduser(paths[index])
            if os.path.isabs(_path):
                paths[index] = os.path.realpath(_path)
            else:
                paths[index] = os.path.realpath(os.path.join(self.build_dir, _path))
        if return_list:
            return paths
        return paths[0]

    def add_artifact(self, artifact_file, properties):
        """
        Register a build artifact to be included in the artifacts manifest.
        """
        self.artifacts[artifact_file] = properties

    def get_source_archive_path(self):
        """
        Create the source archive for use in remote builds or to build the
        source image.
        """
        if not self._source_archive:
            buildignore = os.path.join(self.build_dir, '.buildignore')
            excludes = []
            if os.path.exists(buildignore):
                with open(buildignore, 'r') as _file:
                    excludes = _file.read().splitlines()

            def _exclude_working_dir(tarinfo):
                """
                Filter to exclude results dir from source archive.
                """
                if tarinfo.name == os.path.basename(self.build_results_dir):
                    return None
                for _ex in excludes:
                    if fnmatch.fnmatch(tarinfo.name, _ex):
                        return None
                return tarinfo

            self.log.write('Creating source archive\n')
            _fileobj = None
            try:
                _fileobj = tempfile.NamedTemporaryFile(
                    delete=False,
                    dir=self.get_temp_dir(),
                )
                with tarfile.open(mode='w', fileobj=_fileobj) as tfile:
                    tfile.add(
                        self.build_dir,
                        arcname='',
                        filter=_exclude_working_dir
                    )
                self._source_archive = _fileobj.name
            finally:
                if _fileobj:
                    _fileobj.close()
        return self._source_archive

    def get_source_image(self):
        """
        Get and/or create the base image source containers will be created
        from.
        """
        if not self._source_image:
            self.log.write('Creating source image\n')
            source_archive_path = self.get_source_archive_path()
            inject = {
                source_archive_path: 'source.tar',
                SOURCE_DOCKERFILE: "Dockerfile",
            }
            source_builder = DockerBuilder(
                temp_dir=self.get_temp_dir(),
                inject=inject,
                timeout=self.docker_timeout,
                docker_registry=self.get_docker_registry(),
            )
            exit_code = source_builder.build(
                nocache=True,
                pull=False,
            )
            if exit_code != 0 or not source_builder.image:
                raise BuildRunnerProcessingError(
                    f'Error building source image ({exit_code}), this may be a transient docker'
                    ' error if no output is available above'
                )
            self._source_image = source_builder.image
        return self._source_image

    @property
    def log(self):
        """
        create the log file and open for writing
        """
        if self._log is None:
            try:
                os.makedirs(self.build_results_dir)
            except OSError as exc:
                if exc.errno != errno.EEXIST:
                    sys.stderr.write(f'ERROR: {str(exc)}\n')
                    sys.exit(os.EX_UNAVAILABLE)

            try:
                log_file_path = os.path.join(self.build_results_dir, 'build.log')
                self._log_file = open(log_file_path, 'w')
                self._log = ConsoleLogger(self.colorize_log, self._log_file)

                self.add_artifact(
                    os.path.basename(log_file_path),
                    {'type': 'log'},
                )
            except Exception as exc:  # pylint: disable=broad-except
                sys.stderr.write(f'ERROR: failed to initialize ConsoleLogger: {str(exc)}\n')
                self._log = sys.stderr

        return self._log

    def _write_artifact_manifest(self):
        """
        If we have registered artifacts write the files and associated metadata
        to the artifacts manifest.
        """
        if self.artifacts:
            self.log.write('\nWriting artifact properties\n')
            artifact_manifest = os.path.join(
                self.build_results_dir,
                'artifacts.json',
            )
            # preserve contents of artifacts.json between steps run separately
            if os.path.exists(artifact_manifest):
                with open(artifact_manifest, 'r') as _af:
                    data = json.load(_af, object_pairs_hook=OrderedDict)
                    artifacts = OrderedDict(list(data.items()) + list(self.artifacts.items()))
            else:
                artifacts = self.artifacts

            with open(artifact_manifest, 'w') as _af:
                json.dump(artifacts, _af, indent=2)

    def _exit_message_and_close_log(self, exit_explanation):
        """
        Determine the exit message, output to the log or stdout, close log if
        open.
        """
        if self.exit_code:
            exit_message = '\nBuild ERROR.'
        else:
            exit_message = '\nBuild SUCCESS.'

        if self._log_file:
            try:
                if exit_explanation:
                    self.log.write('\n' + exit_explanation + '\n')
                self.log.write(exit_message + '\n')
            finally:
                # close the log_file
                self._log_file.close()
        else:
            if exit_explanation:
                print(f'\n{exit_explanation}')
            print(exit_message)

    def run(self):  # pylint: disable=too-many-statements,too-many-branches,too-many-locals
        """
        Run the build.
        """
        # reset the exit_code
        self.exit_code = None

        exit_explanation = None
        try:  # pylint: disable=too-many-nested-blocks

            if not os.path.exists(self.build_results_dir):
                # create a new results dir
                os.mkdir(self.build_results_dir)

            self.get_source_archive_path()

            # run each step
            for step_name, step_config in self.run_config['steps'].items():
                if not self.steps_to_run or step_name in self.steps_to_run:
                    image_config = BuildStepRunner.ImageConfig(
                        self.local_images,
                        self.platform
                    )
                    build_step_runner = BuildStepRunner(
                        self,
                        step_name,
                        step_config,
                        image_config
                    )
                    build_step_runner.run()

            self.log.write(
                "\nFinalizing build\n________________________________________\n"
            )

            # see if we should push registered tags to remote registries/repositories
            if self.push:
                self.log.write(
                    'Push requested--pushing generated images/packages to remote registries/repositories\n'
                )
                _docker_client = docker.new_client(timeout=self.docker_timeout)
                for _repo_tag, _insecure_registry in self.repo_tags_to_push:
                    self.log.write(
                        f'\nPushing {_repo_tag}\n'
                    )

                    # Newer Python Docker bindings drop support for the insecure_registry
                    # option.  This test will optionally use it when it's available.
                    push_kwargs = {
                        'stream': True,
                    }
                    if 'insecure_registry' in inspect.getfullargspec(_docker_client.push).args:
                        push_kwargs['insecure_registry'] = _insecure_registry

                    stream = _docker_client.push(_repo_tag, **push_kwargs)
                    previous_status = None
                    for msg_str in stream:
                        for msg in msg_str.decode('utf-8').split("\n"):
                            if not msg:
                                continue
                            msg = json.loads(msg)
                            if 'status' in msg:
                                if msg['status'] == previous_status:
                                    continue
                                self.log.write(msg['status'] + '\n')
                                previous_status = msg['status']
                            elif 'errorDetail' in msg:
                                error_detail = f"Error pushing image: {msg['errorDetail']}\n"
                                self.log.write("\n" + error_detail)
                                self.log.write((
                                    "This could be because you are not "
                                    "authenticated with the given Docker "
                                    "registry (try 'docker login "
                                    "<registry>')\n\n"
                                ))
                                raise BuildRunnerProcessingError(error_detail)
                            else:
                                self.log.write(str(msg) + '\n')

                # Push to pypi repositories
                # Placing the import here avoids the dependency when pypi is not needed
                import twine.commands.upload  # pylint: disable=import-outside-toplevel
                for _, _items in self.pypi_packages.items():
                    twine.commands.upload.upload(_items['upload_settings'], _items['packages'])
            else:
                self.log.write(
                    '\nPush not requested\n'
                )

        except BuildRunnerConfigurationError as brce:
            print('config error')
            exit_explanation = str(brce)
            self.exit_code = os.EX_CONFIG
        except BuildRunnerProcessingError as brpe:
            print('processing error')
            exit_explanation = str(brpe)
            self.exit_code = 1
        except requests.exceptions.ConnectionError as rce:
            print('connection error')
            print(str(rce))
            exit_explanation = (
                "Error communicating with the remote host.\n\tCheck that the "
                "remote docker Daemon is running and/or that the DOCKER_* "
                "environment variables are set correctly.\n\tCheck that the "
                "remote PyPi server information is set correctly."
            )
            self.exit_code = 1

        finally:
            self._write_artifact_manifest()

            _docker_client = docker.new_client(timeout=self.docker_timeout)
            if self.cleanup_images:
                self.log.write(
                    'Removing local copy of generated images\n'
                )
                # cleanup all registered docker images
                # reverse the order of the images since child images would likely come after parent images
                for _image in self.generated_images[::-1]:
                    try:
                        _docker_client.remove_image(
                            _image,
                            noprune=False,
                            force=True,
                        )
                    except Exception as _ex:  # pylint: disable=broad-except
                        self.log.write(
                            f'Error removing image {_image}: {str(_ex)}'
                        )
            else:
                self.log.write(
                    'Keeping generated images\n'
                )

            # cleanup the source image
            if self._source_image:
                self.log.write(
                    f"Destroying source image {self._source_image}\n"
                )
                _docker_client.remove_image(
                    self._source_image,
                    noprune=False,
                    force=True,
                )
            if self._source_archive:
                self.log.write(
                    "Destroying source archive\n"
                )
                os.remove(self._source_archive)

            # remove any temporary files that we created
            for tmp_file in self.tmp_files:
                if os.path.exists(tmp_file):
                    os.remove(tmp_file)

            self._exit_message_and_close_log(exit_explanation)

# Local Variables:
# fill-column: 100
# End:



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import os
import re
from typing import List, Optional, Union

import buildrunner.docker
from buildrunner.errors import (
    BuildRunnerConfigurationError,
    BuildRunnerProcessingError,
)
from buildrunner.steprunner.tasks import BuildStepRunnerTask


def sanitize_tag(tag, log=None):
    """
    :param tag:
    :param log:
    """
    _tag = re.sub(r'[^-_\w.]+', '-', tag.lower())
    if _tag != tag and log:
        log.write(f'Forcing tag to lowercase and removing illegal characters: {tag} => {_tag}\n')
    return _tag


class RepoDefinition:
    """
    Contains the definition for a push repository.
    """

    def __init__(
            self,
            log,
            repository: str,
            tags: Optional[List[str]]=None,
            insecure_registry: Optional[bool]=None,
    ):
        # Force a lower-case repo
        repo_lower = repository.lower()
        if repo_lower != repository:
            log.write(
                f'Forcing repository to lowercase: {repository} => {repo_lower}\n'
            )
        self.repository = repo_lower

        if tags:
            self.tags = [sanitize_tag(tag, log=log) for tag in tags]
        else:
            self.tags = []
        self.insecure_registry = insecure_registry

        # if there is a tag in the repository value, strip it off and add the tag to the list of tags
        tag_index = self.repository.find(":")
        if tag_index > 0:
            tag = self.repository[tag_index + 1:]
            if tag not in self.tags:
                self.tags.append(tag)
            self.repository = self.repository[0:tag_index]


class PushBuildStepRunnerTask(BuildStepRunnerTask):
    """
    Class used to push the resulting image (either from the build task, or if
    there is a run task, the snapshot of the resulting run container) to the
    given registry/repository.
    """
    def _get_repo_definition(self, config: Union[dict, str]) -> RepoDefinition:
        if isinstance(config, dict):
            if 'repository' not in config:
                raise BuildRunnerConfigurationError(
                    'Docker push configuration must at least specify a "repository" attribute'
                )
            return RepoDefinition(
                self.step_runner.log,
                config['repository'],
                config.get('tags'),
                config.get('insecure_registry'),
            )
        return RepoDefinition(self.step_runner.log, config)

    def __init__(self, step_runner, config, commit_only=False):
        super().__init__(step_runner, config)
        self._docker_client = buildrunner.docker.new_client(
            timeout=step_runner.build_runner.docker_timeout,
        )
        self._commit_only = commit_only
        if isinstance(config, list):
            self._repos = [self._get_repo_definition(item) for item in config]
        else:
            self._repos = [self._get_repo_definition(config)]

    def run(self, context):
        # first see if a run task produced an image (via a post-build config)
        if 'run-image' in context:
            image_to_use = context.get('run-image')
        # next see if there was a run task, committing the end state of the
        # container as the image to use
        elif 'run_runner' in context:
            image_to_use = context['run_runner'].commit(self.step_runner.log)
        # finally see if we have an image from a build task
        else:
            image_to_use = context.get('image', None)

        # validate we have an image
        if not image_to_use:
            raise BuildRunnerProcessingError(
                'Cannot find an image to tag/push from a previous task'
            )
        self.step_runner.log.write(
            f'Using image {image_to_use} for tagging\n'
        )

        # Determine internal tag based on source control information and build number
        # Note: do not log sanitization as it would need to happen almost every time
        default_tag = sanitize_tag(self.step_runner.build_runner.build_id)

        # add the image to the list of generated images for potential cleanup
        self.step_runner.build_runner.generated_images.append(image_to_use)

        for repo in self._repos:
            # Always add default tag
            repo.tags.append(default_tag)

            if self._commit_only:
                self.step_runner.log.write(
                    f'Committing resulting image as "{repo.repository}" with tags {", ".join(repo.tags)}.\n'
                )
            else:
                self.step_runner.log.write(
                    f'Preparing resulting image for push to "{repo.repository}" with tags {", ".join(repo.tags)}.\n'
                )

            # Tag the image
            for tag in repo.tags:
                self._docker_client.tag(
                    image_to_use,
                    repo.repository,
                    tag=tag,
                    force=True,
                )
                self.step_runner.build_runner.committed_images.add(f'{repo.repository}:{tag}')

                if not self._commit_only:
                    self.step_runner.build_runner.repo_tags_to_push.append((
                        f"{repo.repository}:{tag}",
                        repo.insecure_registry,
                    ))

            # add image as artifact
            if not self._commit_only:
                self.step_runner.build_runner.add_artifact(
                    os.path.join(self.step_runner.name, image_to_use),
                    {
                        'type': 'docker-image',
                        'docker:image': image_to_use,
                        'docker:repository': repo.repository,
                        'docker:tags': repo.tags,
                    },
                )


class CommitBuildStepRunnerTask(PushBuildStepRunnerTask):
    """
    Class used to commit the resulting image (either from the build task, or if
    there is a run task, the snapshot of the resulting run container) with a
    tag matching the given registry/repository.
    """

    def __init__(self, step_runner, config):
        # Subclasses the push task, just set commit only to true
        super().__init__(step_runner, config, commit_only=True)


# Local Variables:
# fill-column: 100
# End:



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

from collections import OrderedDict
from datetime import datetime
from time import strftime, gmtime
import os
import sys
import uuid
import yaml.resolver
import yaml.scanner
import glob
import hashlib
from typing import Union, Tuple

from buildrunner import BuildRunnerConfigurationError


class OrderedLoader(yaml.Loader):  # pylint: disable=too-many-ancestors
    """
    Custom loader class that preserves dictionary order.
    """
    pass


def construct_mapping(loader, node):
    """
    :param loader:
    :param node:
    """
    loader.flatten_mapping(node)
    return OrderedDict(loader.construct_pairs(node))


OrderedLoader.add_constructor(
    yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
    construct_mapping
)
# Tell YAML how to dump the OrderedDict
yaml.add_representer(
    OrderedDict,
    lambda dumper, data: dumper.represent_dict(iter(data.items())),
)
yaml.add_representer(
    OrderedDict,
    lambda dumper, data: dumper.represent_dict(iter(data.items())),
    Dumper=yaml.SafeDumper
)


class IgnoreAliasesDumper(yaml.Dumper):  # pylint: disable=too-many-ancestors
    """
    Custom dumper class that removes aliases.
    """

    def ignore_aliases(self, data):
        return True


def load_config(stream, cfg_file):
    """
    Load yaml while preserving the order of attributes in maps/dictionaries and
    removing any aliases.
    """
    # run the data through pyyaml again to remove any aliases
    try:
        return yaml.load(
            yaml.dump(
                yaml.load(stream, OrderedLoader),
                default_flow_style=False, Dumper=IgnoreAliasesDumper,
            ),
            Loader=OrderedLoader,
        )
    except yaml.scanner.ScannerError as err:
        raise BuildRunnerConfigurationError(
            f'The {cfg_file} file contains malformed yaml, '
            f'please check the syntax and try again: {err}'
        ) from err


def is_dict(obj):
    """
    Determines whether an object acts like a dict.

    Args:
      obj (object): The object to test.

    Return:
      True if the obj acts like a dict, False otherwise.
    """
    return hasattr(obj, 'keys') and hasattr(obj, '__getitem__')


def epoch_time():
    """Return the current epoch time in integer seconds."""
    return int(
        (datetime.utcnow() - datetime.utcfromtimestamp(0)).total_seconds()
    )


def tempfile(prefix=None, suffix=None, temp_dir='/tmp'):
    """
    Generate a temporary file path within the container.
    """
    name = str(uuid.uuid4())
    if suffix:
        name = name + suffix
    if prefix:
        name = prefix + name

    return os.path.join(temp_dir, name)


def checksum(*files: Tuple[str]) -> str:
    """
    Generate a single SHA1 checksum of the list of files passed in
    """
    if not isinstance(files, tuple):
        print(f"WARNING: TypeError - Input 'files' needs to be a tuple instead of {type(files)}")
        sys.exit()

    blocksize = 2 ** 16
    hasher = hashlib.sha1()

    for filename in sorted(files):
        if not os.path.isfile(filename):
            print(f"WARNING: {filename} does not exist, skipping file.")
            continue

        with open(filename, "rb") as open_file:
            buf = open_file.read(blocksize)
            while len(buf) > 0:
                hasher.update(buf)
                buf = open_file.read(blocksize)

    return hasher.hexdigest()


def hash_sha1(file_name_globs=None):
    """
    Return the sha1 hash from the content of multiple files, represented by a list of globs
    """
    if not file_name_globs:
        file_name_globs = []
    blocksize = 2 ** 16  # 65,536
    hasher = hashlib.sha1()
    for file_name_glob in file_name_globs:
        for file_name in sorted(glob.glob(file_name_glob)):
            try:
                # Use blocksize to ensure python memory isn't too full
                with open(file_name, 'rb') as open_file:
                    buf = open_file.read(blocksize)
                    while len(buf) > 0:
                        hasher.update(buf)
                        buf = open_file.read(blocksize)
            except:
                print(f"WARNING: Error reading file: {file_name}")
    return hasher.hexdigest()


class ConsoleLogger:
    """
    Class used to write decorated output to stdout while also redirecting
    non-decorated output to one or more streams.
    """

    def __init__(self, colorize_log, *streams):
        self.colorize_log = colorize_log
        self.streams = []
        for stream in streams:
            self.streams.append(stream)
        self.stdout = sys.stdout

    def write(self, output: Union[bytes, str], color=None):
        """
        Write the given text to stdout and streams decorating output to stdout
        with color.
        """
        if not isinstance(output, str):
            output = str(output, encoding=sys.stdout.encoding, errors='replace')
        _stdout = output

        # colorize stdout
        if color and self.colorize_log:
            # Colorize stdout
            _stdout = f'[01;3{color}m{_stdout}\033[00;00m'
        self.stdout.write(_stdout)

        # do not colorize output to other streams
        for stream in self.streams:
            try:
                stream.write(output)
            except UnicodeDecodeError as ude:
                stream.write(f"\nERROR writing to log: {str(ude)}\n")

    def flush(self):
        """
        Flush.
        """
        pass


class ContainerLogger:
    """
    Class used to write container specific output to a ConsoleLogger.

    This class is not thread safe, but since each container gets its own that
    is ok.
    """
    BUILD_LOG_COLORS = [3, 4]
    SERVICE_LOG_COLORS = [5, 6, 1, 2]
    LOGGERS = {}

    def __init__(self, console_logger, name, color, timestamps=True):
        self.console_logger = console_logger
        self.name = name
        self.line_prefix = '[' + name + '] '
        self.color = color
        self.timestamps = timestamps
        self._buffer = []

    def _get_timestamp(self):
        if self.timestamps:
            return '[' + strftime("%H:%M:%S", gmtime()) + '] '
        return ''

    def _write_buffer(self):
        """
        Write the contents of the buffer to the log.
        """
        _line = f'{self._get_timestamp()}{self.line_prefix}{"".join(self._buffer)}'
        self._buffer.clear()
        self.console_logger.write(
            _line,
            color=self.color,
        )

    def write(self, output: Union[bytes, str]):
        """
        Write the given output to the log.
        """
        # Ensure that the output is a string
        if not isinstance(output, str):
            output = str(output, encoding=sys.stdout.encoding, errors='replace')

        for char in output:
            self._buffer.append(char)
            if char == '\n':
                self._write_buffer()

    def cleanup(self):
        """
        Flush the buffer.
        """
        self._buffer.append('\n')
        self._write_buffer()

    @classmethod
    def for_build_container(cls, console_logger, name, timestamps=True):
        """
        Return a ContainerLogger for a build container.
        """
        color = cls._cycle_colors(cls.BUILD_LOG_COLORS)
        name_idx = f"{name}{color}"
        if name_idx not in cls.LOGGERS:
            cls.LOGGERS[name_idx] = ContainerLogger(
                console_logger,
                name,
                color,
                timestamps=timestamps,
            )
        return cls.LOGGERS[name_idx]

    @classmethod
    def for_service_container(cls, console_logger, name, timestamps=True):
        """
        Return a ContainerLogger for a service container.
        """
        color = cls._cycle_colors(cls.SERVICE_LOG_COLORS)
        if name not in cls.LOGGERS:
            cls.LOGGERS[name] = ContainerLogger(
                console_logger,
                name,
                color,
                timestamps=timestamps,
            )
        return cls.LOGGERS[name]

    @staticmethod
    def _cycle_colors(colors):
        """
        Cycle through console colors.
        """
        current = colors[0]
        colors[0] = colors[1]
        colors[-1] = current

        return current

# Local Variables:
# fill-column: 100
# End:



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import urllib.parse

from . import github
from . import http
from . import file


def fetch_file(url, config):
    """
    Fetch a file from a provider.
    """

    # pylint: disable=fixme
    # FIXME: the handled checking should be in each handler module (possibly handle_file(parsed_url,
    # config) => bool)

    parsed_url = urllib.parse.urlparse(url)

    if parsed_url.scheme == 'github':
        file_contents = github.fetch_file(parsed_url, config)

    elif parsed_url.scheme == 'file' or (parsed_url.scheme == '' and parsed_url.netloc == ''):
        purl = list(parsed_url)
        purl[0] = 'file'
        parsed_url = urllib.parse.ParseResult(*purl)
        file_contents = file.fetch_file(parsed_url, config)

    elif parsed_url.scheme in ('http', 'https'):
        file_contents = http.fetch_file(parsed_url, config)

    else:
        raise NotImplementedError(f'Unknown fetch backend: {parsed_url.scheme}')

    return file_contents

# Local Variables:
# fill-column: 100
# End:



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""


def fetch_file(parsed_url, config):
    """
    Fetch files using HTTP.
    """

    raise NotImplementedError('Not Implemented: fetch.http.file_fetch')



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import codecs


def fetch_file(parsed_url, config):  # pylint: disable=unused-argument
    """
    Pull files from the local file system.
    """
    with codecs.open(parsed_url.path, 'r', encoding='utf-8') as _file:
        contents = ''.join(_file.readlines())

    return contents



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import os
import traceback
import uuid

from buildrunner.errors import (
    BuildRunnerError,
    BuildRunnerConfigurationError,
    BuildRunnerProcessingError,
)
from buildrunner.steprunner.tasks.build import BuildBuildStepRunnerTask
from buildrunner.steprunner.tasks.push import CommitBuildStepRunnerTask, PushBuildStepRunnerTask
from buildrunner.steprunner.tasks.remote import RemoteBuildStepRunnerTask
from buildrunner.steprunner.tasks.run import RunBuildStepRunnerTask
from buildrunner.steprunner.tasks.pypipush import PypiPushBuildStepRunnerTask

TASK_MAPPINGS = {
    'remote': RemoteBuildStepRunnerTask,
    'build': BuildBuildStepRunnerTask,
    'run': RunBuildStepRunnerTask,
    'push': PushBuildStepRunnerTask,
    'commit': CommitBuildStepRunnerTask,
    'pypi-push': PypiPushBuildStepRunnerTask,
}


class BuildStepRunner:  # pylint: disable=too-many-instance-attributes
    """
    Class used to manage running a build step.
    """

    class ImageConfig:
        """
        An object that captures image-specific configuration
        """

        def __init__(self, local_images=False, platform=None):
            self.local_images = local_images
            self.platform = platform

    def __init__(self, build_runner, step_name, step_config, image_config):
        """
        Constructor.
        """
        local_images = image_config.local_images
        platform = image_config.platform
        self.name = step_name
        self.config = step_config
        self.local_images = local_images
        self.platform=platform

        self.build_runner = build_runner
        self.src_dir = self.build_runner.build_dir
        self.results_dir = os.path.join(
            self.build_runner.build_results_dir,
            self.name,
        )
        if not os.path.exists(self.results_dir):
            os.mkdir(self.results_dir)

        self.log = self.build_runner.log

        # generate a unique step id
        self.id = str(uuid.uuid4())  # pylint: disable=invalid-name

    def run(self):
        """
        Run the build step.
        """
        # create the step results dir
        self.log.write(f'\nRunning step "{self.name}"\n')
        self.log.write(f"{'_' * 40}\n")

        _tasks = []
        _context = {}
        try:
            for _task_name, _task_config in self.config.items():
                self.log.write(f'==> Running step: {self.name}:{_task_name}\n')
                if _task_name in TASK_MAPPINGS:
                    if self.local_images:
                        _task_config['pull'] = False
                    if self.platform:
                        _task_config['platform'] = self.platform
                    _task = TASK_MAPPINGS[_task_name](self, _task_config)
                    _tasks.append(_task)
                    try:
                        _task.run(_context)
                    except BuildRunnerError as err:
                        if not isinstance(_task_config, dict) or not _task_config.get('xfail', False):
                            raise

                        self.log.write(
                            f'Step "{self.name}" failed with exception: {err}\n    '
                            f'Ignoring due to XFAIL\n'
                        )
                else:
                    raise BuildRunnerConfigurationError(
                        f'Step "{self.name}" contains an unknown task "{_task_name}"\n'
                    )
        finally:
            for _task in _tasks:
                try:
                    _task.cleanup(_context)
                except:
                    self.log.write('\nError cleaning up task:\n')
                    traceback.print_exc(file=self.log)



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import os
import base64
import requests

from ..errors import (
    BuildRunnerConfigurationError,
    BuildRunnerProtocolError,
)


def v3_fetch_file(parsed_url, config):
    """
    Fetch files using Github v3 protocol.
    """

    endpoint = config.get('endpoint')
    version = config.get('version')

    username = config.get('username', os.getenv('USER', os.getenv('LOGNAME')))
    if not username:
        raise RuntimeError(f'Failed to look up username to access github {endpoint}')

    auth = (username, config.get('app_token', ''))
    url = '/'.join((
        endpoint,
        version,
        'users',
        username,
    ))
    resp = requests.get(url, auth=auth)
    if resp.status_code != 200:
        raise BuildRunnerProtocolError(f'Failed authenticating {username} on {endpoint}')

    if not parsed_url.path.startswith('/'):
        raise ValueError('URL must begin with "/"')

    fpath = parsed_url.path.split('/')
    ubuild = [endpoint, version, 'repos', fpath[1], fpath[2], 'contents']
    ubuild.extend(fpath[3:])
    url = '/'.join(ubuild)
    resp = requests.get(url, auth=auth)
    if resp.status_code != 200:
        raise BuildRunnerProtocolError(f'Failed fetching URL: {url}')
    json_c = resp.json()

    encoding = json_c.get('encoding')
    enc_contents = json_c.get('content')
    if encoding == 'base64':
        dec_contents = base64.b64decode(enc_contents)
    else:
        raise NotImplementedError(
            f'No implementation to decode {encoding}-encoded contents'
        )

    if resp.encoding == 'utf-8':
        contents = dec_contents.decode('utf-8')
    else:
        raise NotImplementedError(
            f'No implementation to decode {resp.encoding}-encoded contents'
        )

    return contents


def fetch_file(parsed_url, config):
    """
    Fetch a file from Github.
    """

    if parsed_url.scheme != 'github':
        raise ValueError(f'URL scheme must be "github" but is "{parsed_url.github}"')

    ghcfg = config.get('github')
    if not ghcfg:
        raise BuildRunnerConfigurationError('Missing configuration for github in buildrunner.yaml')

    nlcfg = ghcfg.get(parsed_url.netloc)
    if not nlcfg:
        gh_cfgs = ', '.join(ghcfg.keys())
        raise BuildRunnerConfigurationError(
            f'Missing github configuration for {parsed_url.netloc} in buildrunner.yaml'
            f' - known github configurations: {gh_cfgs}'
        )

    ver = nlcfg.get('version')
    # NOTE: potentially the v3_fetch_file() works for other github API versions.
    if ver == 'v3':
        contents = v3_fetch_file(parsed_url, nlcfg)
    else:
        raise NotImplementedError(f'No version support for github API version {ver}')

    return contents



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import glob
import os
import re

import buildrunner.docker
from buildrunner.docker.importer import DockerImporter
from buildrunner.docker.builder import DockerBuilder
from buildrunner.errors import (
    BuildRunnerConfigurationError,
    BuildRunnerProcessingError,
)
from buildrunner.steprunner.tasks import BuildStepRunnerTask
from buildrunner.utils import is_dict


class BuildBuildStepRunnerTask(BuildStepRunnerTask):  # pylint: disable=too-many-instance-attributes
    """
    Class used to manage "build" build tasks.
    """

    DOCKERFILE_IMAGE_REGEX = re.compile(r'^FROM (.*)', re.IGNORECASE)

    def __init__(
            self,
            step_runner,
            config,
            image_to_prepend_to_dockerfile=None,
    ):  # pylint: disable=too-many-statements,too-many-branches,too-many-locals
        super().__init__(step_runner, config)
        self._docker_client = buildrunner.docker.new_client(
            timeout=step_runner.build_runner.docker_timeout,
        )
        self.path = None
        self.dockerfile = None
        self.to_inject = {}
        self.image_to_prepend_to_dockerfile = image_to_prepend_to_dockerfile
        self.nocache = False
        self.cache_from = []
        self.buildargs = {}
        self._import = None
        self.platform = None

        pull_from_config = None
        if not is_dict(self.config):
            self.path = self.config
            self.config = {}
        else:
            self._import = self.config.get('import', self._import)
            self.path = self.config.get('path', self.path)
            self.dockerfile = self.config.get('dockerfile', self.dockerfile)
            self.nocache = self.config.get('no-cache', self.nocache)
            pull_from_config = self.config.get('pull')
            self.platform = self.config.get('platform', self.platform)

            if not is_dict(self.config.get('buildargs', self.buildargs)):
                raise BuildRunnerConfigurationError(
                    f'Step {self.step_runner}:build:buildargs must be a collection/map/dictionary'
                )

            self.buildargs = self.config.get('buildargs', self.buildargs)

            if not isinstance(self.config.get('cache_from', self.cache_from), list):
                raise BuildRunnerConfigurationError(
                    f'Step {self.step_runner}:build:cache_from must be a list'
                )

            self.cache_from = self.config.get('cache_from', self.cache_from)

            for cache_from_image in self.cache_from:
                try:
                    self._docker_client.pull(cache_from_image, platform=self.platform)
                    # If the pull is successful, add the image to be cleaned up at the end of the script
                    self.step_runner.build_runner.generated_images.append(cache_from_image)
                    self.step_runner.log.write(f'Using cache_from image: {cache_from_image}\n')
                except:
                    self.step_runner.log.write(f'WARNING: Unable to pull the cache_from image: {cache_from_image}\n')

            if not is_dict(self.config.get('inject', {})):
                raise BuildRunnerConfigurationError(
                    f'Step {self.step_runner}:build:inject must be a collection/map/dictionary'
                )

            for src_glob, dest_path in self.config.get('inject', {}).items():
                _src_glob = self.step_runner.build_runner.to_abs_path(src_glob)
                xsglob = glob.glob(_src_glob)
                if not xsglob:
                    # Failed to resolve the glob
                    raise BuildRunnerConfigurationError(
                        f'Unable to expand inject glob: {_src_glob}'
                    )
                if len(xsglob) == 1:
                    # Only one source - destination may be directory or filename - check for a trailing
                    # '/' and treat it accordingly.
                    source_file = xsglob[0]
                    if dest_path[-1] == '/' or os.path.split(dest_path)[-1] in ('.', '..'):
                        self.to_inject[source_file] = os.path.normpath(
                            os.path.join(
                                '.',
                                dest_path,
                                os.path.basename(source_file)
                            )
                        )
                    else:
                        self.to_inject[source_file] = os.path.normpath(os.path.join('.', dest_path))
                else:
                    # Multiple sources - destination *must* be a directory - add the source basename
                    # to the dest_dir name.
                    for source_file in xsglob:
                        self.to_inject[source_file] = os.path.normpath(
                            os.path.join(
                                '.',
                                dest_path,
                                os.path.basename(source_file),
                            )
                        )

            if not self._import and not any((
                    self.path, self.dockerfile, self.to_inject
            )):
                raise BuildRunnerConfigurationError(
                    'Docker build context must specify a '
                    '"path", "dockerfile", or "inject" attribute'
                )

        if self.path:
            self.path = self.step_runner.build_runner.to_abs_path(self.path)

        if self.path and not os.path.exists(self.path):
            raise BuildRunnerConfigurationError(
                f'Step {self.step_runner}:build:path:{self.path}: Invalid build context path')

        if not self.dockerfile:
            if self.path:
                path_dockerfile = os.path.join(self.path, 'Dockerfile')
                if os.path.exists(path_dockerfile):
                    self.dockerfile = path_dockerfile
            for src_file, dest_file in self.to_inject.items():
                if os.path.normpath(dest_file) in [
                    'Dockerfile',
                    '/Dockerfile',
                ]:
                    self.dockerfile = src_file

        dockerfile_image = None
        if self.dockerfile:
            _dockerfile_abs_path = self.step_runner.build_runner.to_abs_path(
                self.dockerfile,
            )
            if os.path.exists(_dockerfile_abs_path):
                self.dockerfile = _dockerfile_abs_path
                if self.image_to_prepend_to_dockerfile:
                    # need to load the contents of the Dockerfile so that we
                    # can prepend the image
                    with open(_dockerfile_abs_path, 'r') as _dockerfile:
                        self.dockerfile = _dockerfile.read()

            if self.image_to_prepend_to_dockerfile:
                # prepend the given image to the dockerfile
                self.dockerfile = f'FROM {self.image_to_prepend_to_dockerfile}\n{self.dockerfile}'
                dockerfile_image = self.image_to_prepend_to_dockerfile
            else:
                # Find from image in dockerfile
                match = self.DOCKERFILE_IMAGE_REGEX.match(self.dockerfile)
                if match:
                    dockerfile_image = match.group(1)

        # Set the pull attribute based on configuration or the image itself
        if pull_from_config is not None:
            self.pull = pull_from_config
            self.step_runner.log.write(f'Pulling image was overridden via config to {self.pull}\n')
        elif not dockerfile_image:
            # Default to true if we could not determine the image
            self.pull = True
            self.step_runner.log.write('Could not determine docker image, defaulting pull to true\n')
        else:
            # If the image was previously committed in this run, do not pull by default
            self.pull = dockerfile_image not in self.step_runner.build_runner.committed_images
            self.step_runner.log.write(f'Pull was not specified in configuration, defaulting to {self.pull}\n')

    def run(self, context):
        # 'import' will override other configuration and perform a 'docker
        # import'
        if self._import:
            self.step_runner.log.write(f'  Importing {self._import} as a Docker image\n')
            context['image'] = DockerImporter(
                self._import,
                timeout=self.step_runner.build_runner.docker_timeout,
            ).import_image()
            return

        if not self.dockerfile:
            raise BuildRunnerConfigurationError(
                'Cannot find a Dockerfile in the given path '
                'or inject configurations'
            )

        self.step_runner.log.write('Running docker build\n')
        builder = DockerBuilder(
            self.path,
            inject=self.to_inject,
            dockerfile=self.dockerfile,
            docker_registry=self.step_runner.build_runner.get_docker_registry(),
        )
        try:
            exit_code = builder.build(
                console=self.step_runner.log,
                nocache=self.nocache,
                cache_from=self.cache_from,
                pull=self.pull,
                buildargs=self.buildargs,
                platform=self.platform
            )
            if exit_code != 0 or not builder.image:
                raise BuildRunnerProcessingError('Error building image')
        except Exception as exc:
            self.step_runner.log.write(f'ERROR: {exc}\n')
            raise
        finally:
            builder.cleanup()
        context['image'] = builder.image

# Local Variables:
# fill-column: 100
# End:



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import argparse
from collections import OrderedDict
import logging
import os
import sys


from . import (
    __version__,
    BuildRunner,
    BuildRunnerConfigurationError,
)


PROC_NAME = 'buildrunner'
LOG_NAME = PROC_NAME


LOGLEVEL_NAMES = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
LOGLEVEL_LOOKUP = OrderedDict()
for _ll_name, _ll_val in zip(LOGLEVEL_NAMES, range(len(LOGLEVEL_NAMES))):
    LOGLEVEL_LOOKUP[str(_ll_val)] = _ll_name
    LOGLEVEL_LOOKUP[_ll_name] = _ll_name


def get_logger(loglevel):
    """
    :param loglevel:
    """
    formatter = logging.Formatter('%(asctime)s %(name)-30s %(levelname)-8s %(message)s')
    logger = logging.getLogger(LOG_NAME)
    logger.setLevel(loglevel)

    console_handler = logging.StreamHandler(sys.stderr)
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    return logger


def loglevel_type(string):
    """
    :param string:
    """
    ll_name = string.upper()
    if ll_name not in LOGLEVEL_LOOKUP:
        llevel_values = ', '.join(LOGLEVEL_LOOKUP.keys())
        raise argparse.ArgumentTypeError(
            f'Value "{string}" is not a valid loglevel: {llevel_values}'
        )
    return LOGLEVEL_LOOKUP[ll_name]


def parse_args(argv):
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        prog=argv[0],
        description='buildrunner runs builds defined in buildrunner.yaml'
    )

    parser.add_argument(
        '-c', '--global-config',
        default=None,
        dest='global_config_file',
        help='global configuration file (defaults to "~/.buildrunner.yaml")',
    )

    parser.add_argument(
        '-d', '--directory',
        default=os.getcwd(),
        dest='directory',
        help='build directory (defaults to current working directory)',
    )
    parser.add_argument(
        '-b', '--build-results-directory',
        dest='build_results_dir',
        help='build results directory (defaults to <build-directory>/buildrunner.results)',
    )

    parser.add_argument(
        '-f', '--file',
        default=None,
        dest='config_file',
        help='build configuration file (defaults to "buildrunner.yaml", then "gauntlet.yaml")',
    )

    llevel_values = ', '.join(LOGLEVEL_LOOKUP.keys())
    parser.add_argument(
        '-l', '--loglevel',
        dest='loglevel',
        help=f'Verbosity of output:  Allowed values are {llevel_values}',
        type=loglevel_type,
        default='WARNING',
    )

    parser.add_argument(
        '-n', '--build-number',
        default=None,
        dest='build_number',
        help='build number (defaults to unix/epoch time)',
    )

    parser.add_argument(
        '-t', '--docker-timeout',
        default=600,
        type=int,
        dest='docker_timeout',
        help='docker timeout in seconds, defaults to 600',
    )

    parser.add_argument(
        '--push',
        default=False,
        action='store_true',
        dest='push',
        help='push images to remote registries (without this flag buildrunner simply tags images)',
    )

    parser.add_argument(
        '--keep-images',
        default=False,
        action='store_true',
        dest='keep_images',
        help='keep generated images at the end of the build (images are by default deleted to '
             'prevent clutter on build machines)',
    )

    parser.add_argument(
        '--local-images',
        default=False,
        action='store_true',
        dest='local_images',
        help='Prefer local images rather than fetching remote images. '
             'This can be used for testing images prior to pushing them and is the equivalent '
             'of setting pull to false for every image/build.',
    )

    parser.add_argument(
        '--platform',
        default=None,
        dest='platform',
        type=str,
        help='The platform architecture to pass to the docker daemon when pulling, building, and running images.'
             'For example, "linux/amd64" or "linux/arm64/v8".'
    )

    parser.add_argument(
        '--keep-step-artifacts',
        default=False,
        action='store_true',
        dest='keep_step_artifacts',
        help='keep artifacts generated for each step of the build (step artifacts are by default deleted to '
             'prevent clutter on build machines)',
    )

    parser.add_argument(
        '-s', '--steps',
        default=[],
        dest='steps',
        action='append',
        help='only run the listed steps (use the argument multiple times or specify as comma-delimited)',
    )

    parser.add_argument(
        '--publish-ports',
        default=False,
        action='store_true',
        dest='publish_ports',
        help='publish ports defined on a run step, this should never be used on a build server',
    )

    parser.add_argument(
        '--disable-timestamps',
        default=False,
        action='store_true',
        dest='disable_timestamps',
        help='disables printing of timestamps in the logging output',
    )

    parser.add_argument(
        '--version',
        default=False,
        action='store_true',
        dest='print_version',
        help='print the current buildrunner version and exit',
    )

    parser.add_argument(
        '--no-color',
        default=False,
        action='store_true',
        dest='no_log_color',
        help='disable colors when logging',
    )

    parser.add_argument(
        '--print-generated-files',
        default=False,
        action='store_true',
        dest='print_generated_files',
        help='logs the Jinja generated file contents and stops',
    )

    parser.add_argument(
        '--log-generated-files',
        default=False,
        action='store_true',
        dest='log_generated_files',
        help='logs the Jinja generated file contents',
    )

    args = parser.parse_args(argv[1:])

    # Only absolute directories can do a mount bind
    args.directory = os.path.realpath(args.directory)

    _steps = []
    for _step in args.steps:
        _steps.extend(_step.split(','))
    args.steps = _steps

    return args


def main(argv):
    """Main program execution."""
    args = parse_args(argv)
    logger = get_logger(args.loglevel)
    logger.debug('Startup')

    # are we just printing the version?
    if args.print_version:
        print(__version__)
        return os.EX_OK

    try:
        build_runner = BuildRunner(
            args.directory,
            build_results_dir=args.build_results_dir,
            global_config_file=args.global_config_file,
            run_config_file=args.config_file,
            build_number=args.build_number,
            push=args.push,
            colorize_log=not args.no_log_color,
            cleanup_images=not args.keep_images,
            cleanup_step_artifacts=not args.keep_step_artifacts,
            steps_to_run=args.steps,
            publish_ports=args.publish_ports,
            log_generated_files=(bool(args.log_generated_files or args.print_generated_files)),
            docker_timeout=args.docker_timeout,
            local_images=args.local_images,
            platform=args.platform,
        )

        if not args.print_generated_files:
            build_runner.run()
            if build_runner.exit_code:
                return build_runner.exit_code
    except BuildRunnerConfigurationError as brce:
        print((str(brce)))
        return os.EX_CONFIG
    return os.EX_OK


if __name__ == '__main__':
    sys.exit(main(sys.argv))


# Local Variables:
# fill-column: 100
# End:



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import os

from buildrunner.docker import DOCKER_DEFAULT_DOCKERD_URL


class DockerDaemonProxy:
    """
    Class used to encapsulate Docker daemon information within a container.
    """

    def __init__(self, docker_client, log, docker_registry):
        """
        """
        self.docker_client = docker_client
        self.docker_registry = docker_registry
        self.log = log
        self._daemon_container = None
        self._env = {
            'DOCKER_HOST': DOCKER_DEFAULT_DOCKERD_URL,
        }

    def get_info(self):
        """
        Return a tuple where the first item is the daemon container id and
        the second is a dict of environment variables to be injected into other
        containers providing settings for docker clients to connect to the
        encapsulated daemon.
        """
        return self._daemon_container, self._env

    def start(self):
        """
        Starts a Docker container encapsulating information to connect to the
        current docker daemon.
        """
        _volumes = []
        _binds = {}

        # setup docker env and mounts so that the docker daemon is accessible
        # from within the run container
        for env_name, env_value in os.environ.items():
            if env_name == 'DOCKER_HOST':
                self._env['DOCKER_HOST'] = env_value
            if env_name == 'DOCKER_TLS_VERIFY' and env_value:
                self._env['DOCKER_TLS_VERIFY'] = '1'
            if env_name == 'DOCKER_CERT_PATH':
                if os.path.exists(env_value):
                    _volumes.append('/dockerdaemon/certs')
                    _binds[env_value] = {
                        'bind': '/dockerdaemon/certs',
                        'ro': True,
                    }
                    self._env['DOCKER_CERT_PATH'] = '/dockerdaemon/certs'

        # if DOCKER_HOST is a unix socket we need to mount the socket in the
        # container and adjust the DOCKER_HOST variable accordingly
        docker_host = self._env['DOCKER_HOST']
        if docker_host.startswith('unix://'):
            # need to map the socket as a volume
            local_socket = docker_host.replace('unix://', '')
            if os.path.exists(local_socket):
                _volumes.append('/dockerdaemon/docker.sock')
                _binds[local_socket] = {
                    'bind': '/dockerdaemon/docker.sock',
                    'ro': False,
                }
                self._env['DOCKER_HOST'] = 'unix:///dockerdaemon/docker.sock'

        # create and start the Docker container
        self._daemon_container = self.docker_client.create_container(
            f'{self.docker_registry}/busybox:latest',
            command='/bin/sh',
            volumes=_volumes,
            host_config=self.docker_client.create_host_config(
                binds=_binds
            )
        )['Id']
        self.docker_client.start(self._daemon_container)
        self.log.write(
            f"Created Docker daemon container {self._daemon_container:.10}\n"
        )

    def stop(self):
        """
        Stops the Docker daemon container.
        """
        # kill container
        self.log.write(
            f"Destroying Docker daemon container {self._daemon_container:.10}\n"
        )
        if self._daemon_container:
            self.docker_client.remove_container(
                self._daemon_container,
                force=True,
                v=True,
            )



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

from buildrunner.errors import (
    BuildRunnerConfigurationError,
)
from buildrunner.steprunner.tasks import BuildStepRunnerTask
from buildrunner.utils import is_dict


class PypiPushBuildStepRunnerTask(BuildStepRunnerTask):
    """
    Class used to push the resulting python packages to the given repository.
    """

    def __init__(self, step_runner, config):
        super().__init__(step_runner, config)

        if not self.step_runner.build_runner.push:
            # Was not invoked with ``--push`` so just skip this.  This avoids twine
            # complaining when the push repository is not configured and the user
            # is not even interested in pushing.
            return

        self._repository = None
        self._username = None
        self._password = None
        self._skip_existing = False

        if is_dict(config):
            if 'repository' not in config:
                raise BuildRunnerConfigurationError(
                    'Pypi push configuration must specify a "repository" attribute'
                )
            self._repository = config['repository']

            if 'username' not in config:
                raise BuildRunnerConfigurationError(
                    'Pypi push configuration must specify a "username" attribute'
                )
            self._username = config['username']

            if 'password' not in config:
                raise BuildRunnerConfigurationError(
                    'Pypi push configuration must specify a "password" attribute'
                )
            self._password = config['password']

            if 'skip_existing' not in config:
                self._skip_existing = config['skip_existing']

        else:
            self._repository = config

        if self._repository not in self.step_runner.build_runner.pypi_packages:
            # Importing here avoids twine dependency when it is unnecessary
            import twine.settings  # pylint: disable=import-outside-toplevel
            try:
                if self._username is not None and self._password is not None:
                    upload_settings = twine.settings.Settings(
                        repository_url=self._repository,
                        username=self._username,
                        password=self._password,
                        disable_progress_bar=True,
                        skip_existing=self._skip_existing,
                    )
                else:
                    upload_settings = twine.settings.Settings(
                        repository_name=self._repository,
                        disable_progress_bar=True,
                        skip_existing=self._skip_existing,
                    )
            except twine.exceptions.InvalidConfiguration as twe:
                raise BuildRunnerConfigurationError(
                    f'Pypi is unable to find an entry for "{self._repository}" in your .pypirc.\n'
                ) from twe

            self.step_runner.build_runner.pypi_packages[self._repository] = {
                'upload_settings': upload_settings,
                'packages': [],
            }

    def run(self, context):
        if not self.step_runner.build_runner.push:
            self.step_runner.log.write(
                'Push not requested with "--push": skipping\n'
            )
            return

        self.step_runner.log.write(
            f'Preparing resulting packages for push to "{self._repository}".\n'
        )

        # get python-sdist packages for this step only
        for _artifact, _attributes in self.step_runner.build_runner.artifacts.items():
            if (
                    _artifact.startswith(self.step_runner.name + "/")
                    and _attributes
                    and 'type' in _attributes
                    and _attributes['type'] == "python-sdist"
            ):
                self.step_runner.build_runner.pypi_packages[self._repository]['packages'].append(
                    f"{self.step_runner.build_runner.build_results_dir}/{_artifact}"
                )

# Local Variables:
# fill-column: 100
# End:



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""

import base64
import os.path
import socket
import ssl
from collections import OrderedDict
from os import listdir
from os.path import isfile, join, getmtime
from pathlib import Path
from types import GeneratorType

import docker.errors
import six
from docker.utils import compare_version

from buildrunner.docker import (
    new_client,
    force_remove_container,
    BuildRunnerContainerError,
)
from buildrunner.utils import tempfile


class DockerRunner:
    """
    An object that manages and orchestrates the lifecycle and execution of a
    Docker container.
    """

    class ImageConfig:
        """
        An object that captures image-specific configuration
        """
        def __init__(self, image_name, pull_image=True, platform=None):
            self.image_name = image_name
            self.pull_image = pull_image
            self.platform = platform

    def __init__(self, image_config, dockerd_url=None, log=None):
        image_name = image_config.image_name
        pull_image = image_config.pull_image
        platform = image_config.platform

        self.image_name = image_name.lower()
        self.platform = platform
        if log and self.image_name != image_name:
            log.write(f'Forcing image_name to lowercase: {image_name} => {self.image_name}\n')
        self.docker_client = new_client(
            dockerd_url=dockerd_url,
            # Disable timeouts for running commands
            timeout=0,
        )
        self.container = None
        self.shell = None
        self.committed_image = None
        self.containers = []

        # By default, pull the image.  If the pull_image parameter is
        # set to False, only pull the image if it can't be found locally
        #
        # Pull all images to ensure we get the hashes for intermediate images
        found_image = False
        for image in self.docker_client.images(all=True):
            if image["Id"].startswith("sha256:" + self.image_name) or image["Id"] == self.image_name:
                # If the image name is simply a hash, it refers to an intermediate
                # or imported image.  We don't want to "pull" these, as the hash
                # won't exist as a valid upstream repoistory/image
                found_image = True
                pull_image = False
            else:
                for tag in image['RepoTags'] or []:
                    if tag == self.image_name:
                        found_image = True
                        break
            if found_image:
                # No need to continue once we've found the image
                break

        if pull_image or not found_image:
            if log:
                log.write(f'Pulling image {self.image_name}\n')
            for data in self.docker_client.pull(self.image_name, stream=True, decode=True, platform=self.platform):
                # Unused variable (see comment below about interactive mode)
                _ = data
                if log:
                    log.write('.')
                    # If we implement an interactive mode, this could be used instead
                    # line = data.get('progress', data.get('status')) or '...'
                    # log.write(f'\r{line:<80}')
            if log:
                log.write('\nImage pulled successfully\n')

    def start(
            self,
            shell='/bin/sh',
            working_dir=None,
            name=None,
            volumes=None,
            volumes_from=None,
            links=None,
            ports=None,
            provisioners=None,
            environment=None,
            user=None,
            hostname=None,
            dns=None,
            dns_search=None,
            extra_hosts=None,
            containers=None,
            systemd=None,
            cap_add=None,
            privileged=False,
    ):  # pylint: disable=too-many-arguments,too-many-locals
        """
        Kwargs:
          volumes (dict): mount the local dir (key) to the given container
                          path (value)
        """
        if self.container:
            raise BuildRunnerContainerError('Container already started')
        self.shell = shell

        # save any spawned containers
        if containers:
            self.containers = containers

        # prepare volumes
        _volumes = []
        _binds = {}

        security_opt = None
        command = shell
        if systemd:
            # If we are running in a systemd context,
            # the following 3 settings are necessary to
            # allow services to run.
            volumes["/sys/fs/cgroup"] = "/sys/fs/cgroup:ro"
            security_opt = ["seccomp=unconfined"]
            command = "/usr/sbin/init"

        if volumes:
            for key, value in volumes.items():
                to_bind = value
                _ro = False
                if to_bind.rfind(':') > 0:
                    tokens = to_bind.rsplit(':', 1)
                    to_bind = tokens[0]
                    _ro = tokens[1] == 'ro'
                _volumes.append(to_bind)
                _binds[key] = {
                    'bind': to_bind,
                    'ro': _ro,
                }

        # prepare ports
        _port_list = None
        if ports:
            _port_list = list(ports.keys())

        # check args
        if dns_search and isinstance(dns_search, six.string_types):
            dns_search = dns_search.split(',')

        kwargs = {
            'name': name,
            'command': command,
            'volumes': _volumes,
            'ports': _port_list,
            'stdin_open': True,
            'tty': True,
            'environment': environment,
            'user': user,
            'working_dir': working_dir,
            'hostname': hostname,
            'host_config': self.docker_client.create_host_config(
                binds=_binds,
                links=links,
                port_bindings=ports,
                volumes_from=volumes_from,
                dns=dns,
                dns_search=dns_search,
                extra_hosts=extra_hosts,
                security_opt=security_opt,
                cap_add=cap_add,
                privileged=privileged
            )
        }

        if compare_version('1.10', self.docker_client.api_version) < 0:
            kwargs['dns'] = dns

        # start the container
        self.container = self.docker_client.create_container(self.image_name, **kwargs)
        self.docker_client.start(self.container['Id'])

        # run any supplied provisioners
        if provisioners:
            for provisioner in provisioners:
                try:
                    provisioner.provision(self)
                except Exception as ex:
                    self.cleanup()
                    raise ex

        return self.container['Id']

    def stop(self):
        """
        Stop the backing Docker container.
        """
        if self.container:
            self.docker_client.stop(
                self.container['Id'],
                timeout=0,
            )

    def cleanup(self):
        """
        Cleanup the backing Docker container, stopping it if necessary.
        """
        if self.container:
            for container in self.containers:
                try:
                    force_remove_container(self.docker_client, container)
                except docker.errors.NotFound:
                    try:
                        container_ids = self.docker_client.containers(
                            filters={'label': container},
                            quiet=True
                        )
                        if container_ids:
                            for container_id in container_ids:
                                self.docker_client.remove_container(
                                    container_id['Id'],
                                    force=True,
                                    v=True,
                                )
                        else:
                            print(f'Unable to find docker container with name or label "{container}"')
                    except docker.errors.NotFound:
                        print(f'Unable to find docker container with name or label "{container}"')

            self.docker_client.remove_container(
                self.container['Id'],
                force=True,
                v=True,
            )

        self.container = None

    def restore_caches(self, caches: OrderedDict):  # pylint: disable=too-many-branches,too-many-locals
        """
        Restores caches from the host system to the destination location in the docker container.
        """
        if caches is None or not isinstance(caches, OrderedDict):
            raise TypeError(f"Caches should be type OrderDict instead of {type(caches)}")

        restored_cache_src = set()
        for local_cache_archive_file, docker_path in caches.items():
            if docker_path in restored_cache_src:
                print(f"Destination path {docker_path} has already been matched and restored to the container "
                      f"skipping {local_cache_archive_file} -> {docker_path}")
                continue

            # Check for prefix matching
            if not os.path.exists(local_cache_archive_file):
                cache_dir = os.path.dirname(local_cache_archive_file)

                if not os.path.exists(cache_dir):
                    print(f"Cache directory {cache_dir} does not exist, "
                          f"skipping restore of archive {local_cache_archive_file}")
                    continue

                files = [f for f in listdir(cache_dir) if isfile(join(cache_dir, f))]

                cache_key = Path(local_cache_archive_file).stem

                most_recent_time = 0
                local_cache_archive_match = None
                for file in files:
                    if file.startswith(cache_key):
                        curr_archive_file = join(cache_dir, file)
                        mod_time = getmtime(curr_archive_file)
                        if mod_time > most_recent_time:
                            most_recent_time = mod_time
                            local_cache_archive_match = curr_archive_file

                if local_cache_archive_match is None:
                    print(f"Not able to restore cache {docker_path} since "
                          f"there was no matching prefix for `{local_cache_archive_file}`")
                    continue

                local_cache_archive_file = local_cache_archive_match

            orig_shell = self.shell
            try:
                self.shell = "/bin/sh"
                exit_code = self.run(f"mkdir -p {docker_path}")
                if exit_code:
                    print(f"WARNING: There was an issue creating {docker_path} on the docker container.")

                with open(local_cache_archive_file, 'rb') as data:
                    restored_cache_src.add(docker_path)
                    if not self.docker_client.put_archive(self.container['Id'], docker_path, data):
                        print(f"WARNING: An error occurred when trying to use cache "
                              f"{local_cache_archive_file} at the path {docker_path}")

            except docker.errors.APIError as exception:
                print(f"WARNING: Encountered exception\n{exception}")
            finally:
                self.shell = orig_shell

    def save_caches(self, caches: OrderedDict):
        """
        Saves caches from a source locations in the docker container to locations on the host system as archive file.
        """
        saved_cache_src = set()
        if caches and isinstance(caches, OrderedDict):
            for local_cache_archive_file, docker_path in caches.items():
                if docker_path not in saved_cache_src:
                    saved_cache_src.add(docker_path)
                    print(f"Saving cache `{docker_path}` "
                          f"running on container {self.container['Id']} "
                          f"to local cache `{local_cache_archive_file}`")

                    with open(local_cache_archive_file, 'wb') as file:
                        bits, _ = self.docker_client.get_archive(self.container['Id'], f"{docker_path}/.")
                        for chunk in bits:
                            file.write(chunk)
                else:
                    print(f"The following `{docker_path}` in docker has already been saved. "
                          f"It will not be saved again to `{local_cache_archive_file}`")

    # pylint: disable=too-many-branches,too-many-arguments
    def run(self, cmd, console=None, stream=True, log=None, workdir=None):
        """
        Run the given command in the container.
        """
        # Unused variable
        _ = workdir

        if isinstance(cmd, str):
            cmdv = [self.shell, '-xc', cmd]
        elif hasattr(cmd, 'next') or hasattr(cmd, '__next__') or hasattr(cmd, '__iter__') or \
                isinstance(cmd, GeneratorType):
            cmdv = cmd
        else:
            raise TypeError(f'Unhandled command type: {type(cmd)}:{cmd}')
        # if console is None:
        #    raise Exception('No console!')
        if not self.container:
            raise BuildRunnerContainerError('Container has not been started')
        if not self.shell:
            raise BuildRunnerContainerError(
                'Cannot call run if container cmd not shell'
            )

        if log:
            log.write(f'Executing: {cmdv}\n')

        create_res = self.docker_client.exec_create(
            self.container['Id'],
            cmdv,
            tty=False,
            # workdir=workdir,
        )
        output_buffer = self.docker_client.exec_start(
            create_res,
            stream=stream,
        )
        if isinstance(output_buffer, (bytes, str)):
            if console:
                console.write(output_buffer)
            if log:
                log.write(output_buffer)
        elif hasattr(output_buffer, 'next') or isinstance(output_buffer, GeneratorType):
            try:
                for line in output_buffer:
                    if console:
                        console.write(line)
                    if log:
                        log.write(line)
            except socket.timeout:
                # Ignore timeouts since we check for the exit code anyways at the end
                pass
        else:
            warning = f'WARNING: Unexpected output object: {output_buffer}'
            if console:
                console.write(warning)
            if log:
                log.write(warning)
        inspect_res = self.docker_client.exec_inspect(create_res)
        if 'ExitCode' in inspect_res:
            if inspect_res['ExitCode'] is None:
                raise BuildRunnerContainerError(f'Error running cmd ({cmd}): exit code is None')
            return inspect_res['ExitCode']
        raise BuildRunnerContainerError('Error running cmd: no exit code')

    def run_script(
            self,
            script,
            args='',
            console=None,
    ):
        """
        Run the given script within the container.
        """
        # write temp file with script contents
        script_file_path = tempfile()
        self.run(f'mkdir -p $(dirname {script_file_path})', console=console)
        self.write_to_container_file(script, script_file_path)
        self.run(f'chmod +x {script_file_path}', console=console)

        # execute the script
        return self.run(
            f'{script_file_path} {args}',
            console=console,
        )

    def write_to_container_file(self, content, path):
        """
        Writes contents to the given path within the container.
        """
        # for now, we just take a str
        buf_size = 1024
        for index in range(0, len(content), buf_size):
            self.run(
                f'printf -- "{base64.standard_b64encode(content[index:index + buf_size])}" | base64 --decode >> {path}',
            )

    def _get_status(self):
        """
        Return the status dict for the container.
        """
        status = None
        try:
            status = self.docker_client.inspect_container(
                self.container['Id'],
            )
        except docker.errors.APIError:
            pass
        return status

    def get_ip(self):
        """
        Return the ip address of the running container
        """
        ipaddr = None
        try:
            if self.is_running():
                inspection = self.docker_client.inspect_container(
                    self.container['Id'],
                )
                ipaddr = inspection.get('NetworkSettings', {}).get('IPAddress', None)
        except docker.errors.APIError:
            pass
        return ipaddr

    def is_running(self):
        """
        Return whether the container backed by this Runner is currently
        running.
        """
        status = self._get_status()
        if not status:
            return False
        if 'State' not in status or 'Running' not in status['State']:
            return False
        return status['State']['Running']

    @property
    def exit_code(self):
        """
        Return the exit code of the completed container, or None if it is still
        running.
        """
        status = self._get_status()
        if not status:
            return None
        if 'State' not in status or 'ExitCode' not in status['State']:
            return None
        return status['State']['ExitCode']

    def attach_until_finished(self, stream=None):
        """
        Attach to the container, writing output to the given log stream until
        the container exits.
        """
        docker_socket: socket.SocketIO = self.docker_client.attach_socket(
            self.container['Id'],
        )
        running = self.is_running()
        while running:
            try:
                for line in docker_socket:
                    if stream:
                        stream.write(line)
            except socket.timeout:
                pass
            except ssl.SSLError as ssle:
                if 'The read operation timed out' not in str(ssle):
                    raise
            running = self.is_running()

    def commit(self, stream):
        """
        Commit the ending state of the container as an image, returning the
        image id.
        """
        if self.committed_image:
            return self.committed_image
        if not self.container:
            raise BuildRunnerContainerError('Container not started')
        if self.is_running():
            raise BuildRunnerContainerError('Container is still running')
        stream.write(
            f"Committing build container {self.container['Id']:.10} as an image...\n"
        )
        self.committed_image = self.docker_client.commit(
            self.container['Id'],
        )['Id']
        stream.write(
            f'Resulting build container image: {self.committed_image:.10}\n'
        )
        return self.committed_image



"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""
# pylint: disable=too-many-lines

from collections import OrderedDict
import grp
import os
import pwd
import threading
import time
import uuid

import buildrunner.docker
from buildrunner.docker.daemon import DockerDaemonProxy
from buildrunner.docker.runner import DockerRunner
from buildrunner.errors import (
    BuildRunnerConfigurationError,
    BuildRunnerProcessingError,
)
from buildrunner.provisioners import create_provisioners
from buildrunner.sshagent import DockerSSHAgentProxy
from buildrunner.steprunner.tasks import BuildStepRunnerTask
from buildrunner.steprunner.tasks.build import BuildBuildStepRunnerTask
from buildrunner.utils import ContainerLogger, is_dict

DEFAULT_SHELL = '/bin/sh'
SOURCE_VOLUME_MOUNT = '/source'
ARTIFACTS_VOLUME_MOUNT = '/artifacts'
FILE_INFO_DELIMITER = '~!~'


class RunBuildStepRunnerTask(BuildStepRunnerTask):
    """
    Class used to manage "run" build tasks.
    """

    # Lightweight docker image for artifact management
    ARTIFACT_LISTER_DOCKER_IMAGE = 'ubuntu:19.04'
    # Lightweight docker image for running nc
    NC_DOCKER_IMAGE = 'subfuzion/netcat:latest'

    # Default to 10 min when waiting for ports to be opened
    WAIT_FOR_DEFAULT_TIMEOUT = 600

    TAR_COMPRESSION_ARG = {
        'gz': '--gzip',
        'bz2': '--bzip2',
        'xz': '--xz',
        'lzma': '--lzma',
        'lzip': '--lzip',
        'lzop': '--lzop',
        'z': '-Z',
    }

    def __init__(self, step_runner, config):
        super().__init__(step_runner, config)
        self._docker_client = buildrunner.docker.new_client(
            timeout=step_runner.build_runner.docker_timeout,
        )
        self._source_container = None
        self._service_runners = OrderedDict()
        self._service_links = {}
        self._sshagent = None
        self._dockerdaemonproxy = None
        self.runner = None

    def _get_source_container(self):
        """
        Get (creating the container if necessary) the container id of the
        source container.
        """
        if not self._source_container:
            self._source_container = self._docker_client.create_container(
                self.step_runner.build_runner.get_source_image(),
                command='/bin/sh',
            )['Id']
            self._docker_client.start(
                self._source_container,
            )
            self.step_runner.log.write(
                f'Created source container {self._source_container:.10}\n'
            )
        return self._source_container

    def _process_volumes_from(self, volumes_from):
        """
        Translate the volumes_from configuration to the appropriate service
        container ids.
        """
        _volumes_from = []
        for sc_vf in volumes_from:
            volumes_from_definition = sc_vf.rsplit(':')
            service_container = volumes_from_definition[0]
            volume_option = None
            if len(volumes_from_definition) > 1:
                volume_option = volumes_from_definition[1]
            if service_container not in list(self._service_links.values()):
                raise BuildRunnerConfigurationError(
                    f'"volumes_from" configuration "{sc_vf}" does not reference a valid service container\n'
                )
            for container, service in self._service_links.items():
                if service == service_container:
                    if volume_option:
                        _volumes_from.append(
                            f"{container}:{volume_option}",
                        )
                    else:
                        _volumes_from.append(container)
                    break
        return _volumes_from

    def _retrieve_artifacts(self, console=None):  # pylint: disable=too-many-locals
        """
        Gather artifacts from the build container and place in the
        step-specific results dir.

        This function actually spins up a separate docker container that mounts
        the build container's source directory, queries for artifact patterns,
        then copies to the step-specific results directory (which is mounted as
        a host volume). It also appropriately sets file ownership of the
        archived artifacts to the user running the buildrunner process.
        """
        self.step_runner.log.write('Gathering artifacts\n')
        if not self.config['artifacts']:
            return

        # use a small busybox image to list the files matching the glob
        artifact_lister = None
        try:
            image_config = DockerRunner.ImageConfig(
                f'{self.step_runner.build_runner.get_docker_registry()}/{self.ARTIFACT_LISTER_DOCKER_IMAGE}',
                pull_image=False,
            )
            artifact_lister = DockerRunner(
                image_config,
                log=self.step_runner.log,
            )
            # NOTE: see if we can use archive commands to eliminate the need for
            #       the /stepresults volume when we can move to api v1.20
            artifact_lister.start(
                volumes_from=[self._get_source_container()],
                volumes={
                    self.step_runner.results_dir: '/stepresults',
                },
                working_dir=SOURCE_VOLUME_MOUNT,
                shell='/bin/sh',
            )

            for pattern, properties in self.config['artifacts'].items():
                # query files for each artifacts pattern, capturing the output
                # for parsing
                stat_output_file = f"{str(uuid.uuid4())}.out"
                stat_output_file_local = os.path.join(
                    self.step_runner.results_dir,
                    stat_output_file,
                )
                exit_code = artifact_lister.run(
                    f'stat -c "%n{FILE_INFO_DELIMITER}%F" {pattern} >/stepresults/{stat_output_file}',
                    console=console,
                    stream=True,
                    log=self.step_runner.log,
                )

                # if the command was successful we found something
                if exit_code == 0:
                    with open(stat_output_file_local, 'r') as output_fd:
                        output = output_fd.read()
                    artifact_files = [
                        af.strip() for af in output.split('\n')
                    ]
                    for art_info in artifact_files:
                        if not art_info or FILE_INFO_DELIMITER not in art_info:
                            continue
                        artifact_file, file_type = art_info.split(
                            FILE_INFO_DELIMITER,
                        )

                        # check if the file is directory, to copy recursive
                        is_dir = file_type.strip() == 'directory'

                        if is_dir:
                            # directory => recursive copy
                            self._archive_dir(
                                artifact_lister,
                                properties,
                                artifact_file,
                            )

                        else:
                            output_file_name = os.path.basename(artifact_file)
                            new_artifact_file = '/stepresults/' + output_file_name
                            archive_command = (
                                'cp',
                                '-L',
                                artifact_file,
                                new_artifact_file,
                            )
                            self._archive_file(
                                artifact_lister,
                                'file',
                                output_file_name,
                                archive_command,
                                artifact_file,
                                new_artifact_file,
                                output_file_name,
                                properties,
                            )

                # remove the stat output file
                if os.path.exists(stat_output_file_local):
                    os.remove(stat_output_file_local)

        finally:
            if artifact_lister:
                # make sure the current user/group ids of our
                # process are set as the owner of the files
                exit_code = artifact_lister.run(
                    f'chown -R {int(os.getuid())}:{int(os.getgid())} /stepresults',
                    console=console,
                    log=self.step_runner.log,
                )
                if exit_code != 0:
                    raise Exception(
                        "Error gathering artifacts--unable to change ownership"
                    )

                artifact_lister.cleanup()

    def _archive_dir(
            self,
            artifact_lister,
            properties,
            artifact_file,
    ):  # pylint: disable=too-many-locals
        """
        Archive the given directory.
        """
        if properties and properties.get('format', None) == 'uncompressed':
            # recursively find all files in dir and add
            # each one, passing any properties
            find_output_file = f"{str(uuid.uuid4())}.out"
            find_output_file_local = os.path.join(
                self.step_runner.results_dir,
                find_output_file,
            )
            find_exit_code = artifact_lister.run(
                f'find {artifact_file} -type f >/stepresults/{find_output_file}',
                stream=False,
                log=self.step_runner.log,
            )
            if find_exit_code == 0:
                with open(find_output_file_local, 'r') as output_fd:
                    output = output_fd.read()
                for _file in [_f.strip() for _f in output.split('\n')]:
                    if not _file:
                        continue

                    output_file_name = _file
                    new_artifact_file = '/stepresults/' + output_file_name
                    _dir_name = os.path.dirname(_file)
                    archive_command = (
                        'mkdir',
                        '-p',
                        '/stepresults/' + _dir_name,
                    )
                    exit_code = artifact_lister.run(
                        archive_command,
                        log=self.step_runner.log,
                    )
                    if exit_code != 0:
                        raise Exception(
                            f"Error gathering artifact {artifact_file}",
                        )
                    archive_command = (
                        'cp',
                        '-r',
                        _file,
                        new_artifact_file,
                    )
                    self._archive_file(
                        artifact_lister,
                        'file',
                        _file,
                        archive_command,
                        _file,
                        new_artifact_file,
                        output_file_name,
                        properties,
                    )

                # remove the find output file
                if os.path.exists(find_output_file_local):
                    os.remove(find_output_file_local)

        else:
            filename = os.path.basename(artifact_file)
            arch_props = {
                'name': filename,
                'compression': 'gz',
                'type': 'tar',
            }
            if properties:
                arch_props.update(properties.get('archive', {}))

            workdir = None
            if arch_props['type'] == 'tar':
                suffix = arch_props.get('suffix', f'.{arch_props["type"]}.{arch_props["compression"]}')
                output_file_name = arch_props['name'] + suffix
                new_artifact_file = '/stepresults/' + output_file_name
                archive_command = [
                    'tar',
                    self.TAR_COMPRESSION_ARG.get(arch_props['compression'], '--auto-compress'),
                    '--xform', f"s|^{filename}|{arch_props['name']}|",
                    '-cv',
                ]
                if os.path.dirname(artifact_file):
                    archive_command.extend((
                        '-C', os.path.dirname(artifact_file),
                    ))
                archive_command.extend((
                    '-f', new_artifact_file,
                    filename,
                ))

            elif arch_props['type'] == 'zip':
                output_file_name = f'{arch_props["name"]}.{arch_props["type"]}'
                new_artifact_file = '/stepresults/' + output_file_name
                archive_command = [
                    'zip',
                    new_artifact_file,
                    artifact_file,
                ]

            new_properties = dict()
            if properties:
                new_properties.update(properties)
            new_properties[
                'buildrunner.compressed.directory'
            ] = 'true'
            self._archive_file(
                artifact_lister,
                'directory',
                filename,
                archive_command,
                artifact_file,
                new_artifact_file,
                output_file_name,
                new_properties,
                workdir=workdir,
            )

    def _archive_file(
            self,
            artifact_lister,
            file_type,
            filename,
            archive_command,
            artifact_file,
            new_artifact_file,
            output_file_name,
            properties,
            workdir=None,
    ):  # pylint: disable=too-many-arguments
        """
        Archive the given file.
        """
        # Unused arg
        _ = new_artifact_file

        self.step_runner.log.write(
            f'- found {file_type} {filename}\n'
        )

        exit_code = artifact_lister.run(
            archive_command,
            log=self.step_runner.log,
            workdir=workdir,
        )
        if exit_code != 0:
            raise Exception(
                f"Error gathering artifact {artifact_file}",
            )

        # register the artifact with the run controller
        self.step_runner.build_runner.add_artifact(
            os.path.join(
                self.step_runner.name,
                output_file_name,
            ),
            properties,
        )

    # pylint: disable=too-many-statements,too-many-branches,too-many-locals
    def _start_service_container(self, name, config):
        """
        Start a service container.
        """
        # validate that we have an 'image' or 'build' config
        if not ('image' in config or 'build' in config):
            raise BuildRunnerConfigurationError(
                f'Step "{self.step_runner.name}", service "{name}" must specify an image or docker build context'
            )
        if 'image' in config and 'build' in config:
            raise BuildRunnerConfigurationError(
                f'Step "{self.step_runner.name}", service "{name}" must specify '
                f'either an image or docker build context, not both'
            )

        _image = None
        # see if we need to build an image
        if 'build' in config:
            build_image_task = BuildBuildStepRunnerTask(
                self.step_runner,
                config['build'],
            )
            _build_context = {}
            build_image_task.run(_build_context)
            _image = _build_context.get('image', None)

        if 'image' in config:
            _image = config['image']
        assert _image

        self.step_runner.log.write(
            f'Creating service container "{name}" from image "{_image}\"\n'
        )
        service_logger = ContainerLogger.for_service_container(
            self.step_runner.log,
            name,
            timestamps=not self.step_runner.build_runner.disable_timestamps,
        )

        # setup custom env variables
        _env = dict(self.step_runner.build_runner.env)
        _env['BUILDRUNNER_INVOKE_USER'] = pwd.getpwuid(os.getuid())
        _env['BUILDRUNNER_INVOKE_UID'] = os.getuid()
        _env['BUILDRUNNER_INVOKE_GROUP'] = grp.getgrgid(os.getgid())
        _env['BUILDRUNNER_INVOKE_GID'] = os.getgid()

        # do we need to change to a given dir when running
        # a cmd or script?
        _cwd = None
        if 'cwd' in config:
            _cwd = config['cwd']

        # need to expose any ports?
        _ports = None
        if 'ports' in config:
            _ports = config['ports']

        # default to a container that runs the default
        # image command
        _shell = None

        # do we need to run an explicit cmd?
        if 'cmd' in config:
            # if so we need to run the cmd within a default
            # shell--specify it here
            _shell = DEFAULT_SHELL

        # if a shell is specified use it
        if 'shell' in config:
            _shell = config['shell']

        # see if there are any provisioners defined
        _provisioners = None
        if 'provisioners' in config:
            _provisioners = create_provisioners(
                config['provisioners'],
                service_logger,
            )

        # determine if a user is specified
        _user = None
        if 'user' in config:
            _user = config['user']

        # determine if a hostname is specified
        _hostname = None
        if 'hostname' in config:
            _hostname = config['hostname']

        # determine if a dns host is specified
        _dns = None
        if 'dns' in config:
            # If the dns host is set to a string and that string is a reference to a running service
            # container, pull the ip address out of the service container.
            _dns = [self._resolve_service_ip(config['dns'])]

        # determine if a dns_search domain is specified
        _dns_search = None
        if 'dns_search' in config:
            _dns_search = config['dns_search']

        # determine if any extra hosts were provided
        _extra_hosts = None
        if 'extra_hosts' in config:
            _extra_hosts = {}
            for extra_host, extra_host_ip in config['extra_hosts'].items():
                _extra_hosts[extra_host] = self._resolve_service_ip(extra_host_ip)

        # set service specific environment variables
        _env['BUILDRUNNER_STEP_ID'] = self.step_runner.id
        _env['BUILDRUNNER_STEP_NAME'] = self.step_runner.name
        if 'env' in config:
            for key, value in config['env'].items():
                _env[key] = value

        # make buildrunner aware of spawned containers
        _containers = None
        if 'containers' in config:
            _containers = config['containers']

        # wait for ports on this container to be listening
        # before moving on
        _wait_for = []
        if 'wait_for' in config:
            _wait_for = config['wait_for']

        _volumes_from = [self._get_source_container()]

        # attach the ssh agent to the service container
        if self._sshagent and config.get('inject-ssh-agent', False):
            ssh_container, ssh_env = self._sshagent.get_info()
            if ssh_container:
                _volumes_from.append(ssh_container)
            if ssh_env:
                for _var, _val in ssh_env.items():
                    _env[_var] = _val

        # attach the docker daemon container
        daemon_container, daemon_env = self._dockerdaemonproxy.get_info()
        if daemon_container:
            _volumes_from.append(daemon_container)
        if daemon_env:
            for _var, _val in daemon_env.items():
                _env[_var] = _val

        # see if we need to map any service container volumes
        if 'volumes_from' in config:
            _volumes_from.extend(self._process_volumes_from(
                config['volumes_from'],
            ))

        _volumes = {
            self.step_runner.build_runner.build_results_dir: \
                ARTIFACTS_VOLUME_MOUNT + ':ro',
        }
        if 'files' in config:
            for f_alias, f_path in config['files'].items():
                # lookup file from alias
                f_local = self.step_runner.build_runner.get_local_files_from_alias(
                    f_alias,
                )
                if not f_local or not os.path.exists(f_local):
                    raise BuildRunnerConfigurationError(
                        f"Cannot find valid local file for alias '{f_alias}'"
                    )

                if f_path[-3:] not in [':ro', ':rw']:
                    f_path = f_path + ':ro'

                _volumes[f_local] = f_path

                service_logger.write(
                    f"Mounting {f_local} -> {f_path}\n"
                )

        # instantiate and start the runner
        image_config = DockerRunner.ImageConfig(
            _image,
            config.get('pull', True),
            config.get('platform', None),
        )
        service_runner = DockerRunner(
            image_config,
            log=service_logger,
        )
        self._service_runners[name] = service_runner
        cont_name = self.step_runner.id + '-' + name
        service_container_id = service_runner.start(
            name=cont_name,
            volumes=_volumes,
            volumes_from=_volumes_from,
            ports=_ports,
            links=self._service_links,
            shell=_shell,
            provisioners=_provisioners,
            environment=_env,
            user=_user,
            hostname=_hostname,
            dns=_dns,
            dns_search=_dns_search,
            extra_hosts=_extra_hosts,
            working_dir=_cwd,
            containers=_containers,
            systemd=self.is_systemd(config, _image, service_logger)

        )
        self._service_links[cont_name] = name

        def attach_to_service():
            """Function to attach to service in a separate thread."""
            # if specified, run a command
            if 'cmd' in config:
                exit_code = service_runner.run(
                    config['cmd'],
                    console=service_logger,
                    # log=self.step_runner.log,
                )
                if exit_code != 0:
                    service_logger.write(
                        f'Service command "{config["cmd"]}" exited with code {exit_code}\n'
                    )
            else:
                service_runner.attach_until_finished(service_logger)
            service_logger.cleanup()

        # Attach to the container in a separate thread
        service_management_thread = threading.Thread(
            name=f"{self.step_runner.name}--{name}",
            target=attach_to_service,
        )
        service_management_thread.daemon = True
        service_management_thread.start()

        # wait for listening ports on this container
        for wait_for_data in _wait_for:
            self.wait(cont_name, wait_for_data)

        self.step_runner.log.write(
            f'Started service container "{name}" ({service_container_id:.10})\n'
        )

    def wait(self, name, wait_for_data):
        """
        Wait for listening port on named container
        """
        ipaddr = self._docker_client.inspect_container(name)['NetworkSettings']['IPAddress']
        socket_open = False

        if isinstance(wait_for_data, dict):
            port = wait_for_data.get('port')
            if not port:
                raise BuildRunnerProcessingError('Configuration error, wait_for must include a port')
            timeout = wait_for_data.get('timeout', self.WAIT_FOR_DEFAULT_TIMEOUT)
            if not isinstance(timeout, int):
                timeout = int(timeout)
        else:
            port = wait_for_data
            timeout = self.WAIT_FOR_DEFAULT_TIMEOUT

        start_time = time.time()
        while not socket_open:
            time_spent = int(time.time() - start_time)
            self.step_runner.log.write(
                f"Waiting for port {port} to be listening for connections in container {name} "
                f"with IP address {ipaddr} ({time_spent}/{timeout} seconds elapsed)\n"
            )

            # check that the container is still available
            container = self._docker_client.inspect_container(name)
            container_status = container.get('State', {}).get('Status')
            if container_status not in ['created', 'running']:
                raise BuildRunnerProcessingError(
                    f'Unable to wait for a service port {port} to be ready, the container'
                    f' {name} status is {container_status}'
                )

            # Use a small nc image to test if the port is open from within the docker network
            # Linux can talk to containers directly, but mac and other OSes cannot
            # See https://github.com/docker/for-mac/issues/155 for more info for mac
            nc_tester = None
            try:
                image_config = DockerRunner.ImageConfig(
                    f'{self.step_runner.build_runner.get_docker_registry()}/{self.NC_DOCKER_IMAGE}',
                    pull_image=False,
                )
                nc_tester = DockerRunner(
                    image_config,
                    # Do not log anything from this container
                    log=None,
                )
                nc_tester.start(
                    # The shell is the command
                    shell=f'-n -z {ipaddr} {port}',
                )

                nc_tester.attach_until_finished()
                exit_code = nc_tester.exit_code
                socket_open = exit_code is not None and not exit_code
            finally:
                if nc_tester:
                    nc_tester.cleanup()

            if not socket_open:
                # Make sure we have not yet timed out
                if time_spent > timeout:
                    raise BuildRunnerProcessingError(
                        f'Timed out waiting for port {port} to be opened in container {name} '
                        f'with IP address {ipaddr} after {timeout} seconds'
                    )
                time.sleep(1)

        self.step_runner.log.write(f"Port {int(port)} is listening in container {name} with IP address {ipaddr}\n")

    def _resolve_service_ip(self, service_name):
        """
        If service_name represents a running service, return it's IP address.
        Otherwise, return the service_name
        """
        rval = service_name
        if isinstance(service_name, str) and service_name in self._service_runners:
            ipaddr = self._service_runners[service_name].get_ip()
            if ipaddr is not None:
                rval = ipaddr
        return rval

    def run(self, context):  # pylint: disable=too-many-statements,too-many-branches,too-many-locals
        # Set pull attribute based on configured image
        # Default to configuration
        if self.config.get('pull') is not None:
            pull_image = self.config.get('pull')
            self.step_runner.log.write(f'Pulling image was overridden via config to {pull_image}\n')
        else:
            # Default to pulling, but check if image was committed by buildrunner and set to false if so
            pull_image = True
            config_image = self.config.get('image')
            if config_image:
                pull_image = config_image not in self.step_runner.build_runner.committed_images
                self.step_runner.log.write(f'Pull was not specified in configuration, defaulting to {pull_image}\n')

        _run_image = self.config.get('image', context.get('image', None))
        if not _run_image:
            raise BuildRunnerConfigurationError(
                'Docker run context must specify a "image" attribute or '
                'be preceded by a build context'
            )
        _lrun_image = _run_image.lower()
        if _lrun_image != _run_image:
            self.step_runner.log.write(
                f'Forcing image name to lowercase: {_run_image} => {_lrun_image}\n'
            )
            _run_image = _lrun_image

        self.step_runner.log.write(
            f'Creating build container from image "{_run_image}\"\n'
        )
        container_logger = ContainerLogger.for_build_container(
            self.step_runner.log,
            self.step_runner.name,
            timestamps=not self.step_runner.build_runner.disable_timestamps,
        )
        container_meta_logger = ContainerLogger.for_build_container(
            self.step_runner.log,
            self.step_runner.name,
            timestamps=not self.step_runner.build_runner.disable_timestamps,
        )

        # container defaults
        _source_container = self._get_source_container()
        _container_name = str(uuid.uuid4())
        _env_defaults = dict(self.step_runner.build_runner.env)
        _env_defaults.update({
            'BUILDRUNNER_SOURCE_CONTAINER': _source_container,
            'BUILDRUNNER_BUILD_CONTAINER': _container_name,
        })
        container_args = {
            'name': _container_name,
            'hostname': None,
            'working_dir': SOURCE_VOLUME_MOUNT,
            'shell': None,
            'user': None,
            'provisioners': None,
            'dns': None,
            'dns_search': None,
            'extra_hosts': None,
            'environment': _env_defaults,
            'containers': None,
            'ports': None,
            'volumes_from': [_source_container],
            'volumes': {
                self.step_runner.build_runner.build_results_dir: (
                        ARTIFACTS_VOLUME_MOUNT + ':ro'
                ),
            },
            'cap_add': None,
            'privileged': None
        }
        caches = OrderedDict()

        # see if we need to inject ssh keys
        if 'ssh-keys' in self.config:
            self._sshagent = DockerSSHAgentProxy(
                self._docker_client,
                self.step_runner.log,
                self.step_runner.build_runner.get_docker_registry(),
            )
            self._sshagent.start(
                self.step_runner.build_runner.get_ssh_keys_from_aliases(
                    self.config['ssh-keys'],
                )
            )

        # start the docker daemon proxy
        self._dockerdaemonproxy = DockerDaemonProxy(
            self._docker_client,
            self.step_runner.log,
            self.step_runner.build_runner.get_docker_registry(),
        )
        self._dockerdaemonproxy.start()

        # start any service containers
        if 'services' in self.config:
            for _name, _config in self.config['services'].items():
                self._start_service_container(_name, _config)

        # determine if there is a command to run
        _cmds = []
        if 'cmd' in self.config:
            container_args['shell'] = DEFAULT_SHELL
            _cmds.append(self.config['cmd'])
        if 'cmds' in self.config:
            container_args['shell'] = DEFAULT_SHELL
            _cmds.extend(self.config['cmds'])

        if 'provisioners' in self.config:
            container_args['shell'] = DEFAULT_SHELL
            container_args['provisioners'] = create_provisioners(
                self.config['provisioners'],
                container_logger,
            )

        # if a shell is specified use it
        if 'shell' in self.config:
            container_args['shell'] = self.config['shell']

        # determine the working dir the build should be run in
        if 'cwd' in self.config:
            container_args['working_dir'] = self.config['cwd']

        # determine if a user is specified
        if 'user' in self.config:
            container_args['user'] = self.config['user']

        # determine if a hostname is specified
        if 'hostname' in self.config:
            container_args['hostname'] = self.config['hostname']

        # determine if a dns host is specified
        if 'dns' in self.config:
            # If the dns host is set to a string and that string is a reference to a running service
            # container, pull the ip address out of the service container.
            container_args['dns'] = [self._resolve_service_ip(self.config['dns'])]

        # determine if a dns_search domain is specified
        if 'dns_search' in self.config:
            container_args['dns_search'] = self.config['dns_search']

        # determine additional hosts to add
        if 'extra_hosts' in self.config:
            extra_hosts = {}
            for extra_host, extra_host_ip in self.config['extra_hosts'].items():
                extra_hosts[extra_host] = self._resolve_service_ip(extra_host_ip)
            container_args['extra_hosts'] = extra_hosts

        # set step specific environment variables
        container_args['environment']['BUILDRUNNER_STEP_ID'] = self.step_runner.id
        container_args['environment']['BUILDRUNNER_STEP_NAME'] = self.step_runner.name
        if 'env' in self.config:
            for key, value in self.config['env'].items():
                container_args['environment'][key] = value

        # make buildrunner aware of spawned containers
        if 'containers' in self.config:
            container_args['containers'] = self.config['containers']

        # see if we need to map any service container volumes
        if 'volumes_from' in self.config:
            container_args['volumes_from'].extend(self._process_volumes_from(
                self.config['volumes_from'],
            ))

        # see if we need to attach to a sshagent container
        if self._sshagent:
            ssh_container, ssh_env = self._sshagent.get_info()
            if ssh_container:
                container_args['volumes_from'].append(ssh_container)
            if ssh_env:
                for _var, _val in ssh_env.items():
                    container_args['environment'][_var] = _val

        # attach the docker daemon container
        daemon_container, daemon_env = self._dockerdaemonproxy.get_info()
        if daemon_container:
            container_args['volumes_from'].append(daemon_container)
        if daemon_env:
            for _var, _val in daemon_env.items():
                container_args['environment'][_var] = _val

        # see if we need to inject any files
        if 'files' in self.config:
            for f_alias, f_path in self.config['files'].items():
                # lookup file from alias
                f_local = self.step_runner.build_runner.get_local_files_from_alias(
                    f_alias,
                )
                if not f_local:
                    f_local = os.path.realpath(os.path.join(
                        self.step_runner.build_runner.build_dir,
                        f_alias
                    ))
                    if (
                            f_local != self.step_runner.build_runner.build_dir
                            and not f_local.startswith(self.step_runner.build_runner.build_dir + os.path.sep)
                    ):
                        raise BuildRunnerConfigurationError(
                            f'Mount path of "{f_alias}" attempts to step out of '
                            f'source directory "{self.step_runner.build_runner.build_dir}"'
                        )

                    if not os.path.exists(f_local):
                        raise BuildRunnerConfigurationError(
                            f"Cannot find valid alias for files entry '{f_alias}' nor path at '{f_local}'"
                        )

                if f_path[-3:] not in [':ro', ':rw']:
                    f_path = f_path + ':ro'

                container_args['volumes'][f_local] = f_path

                container_meta_logger.write(
                    f"Mounting {f_local} -> {f_path}\n"
                )

        if 'caches' in self.config:
            for key, value in self.config['caches'].items():

                if isinstance(value, str):
                    # get the cache location from the main BuildRunner class
                    cache_archive_file = self.step_runner.build_runner.get_cache_archive_file(
                        cache_name=key,
                        project_name=self.step_runner.build_runner.vcs.name
                    )
                    caches[cache_archive_file] = value
                    container_meta_logger.write(
                        f"Copying local cache `{cache_archive_file}` -> docker path `{value}`\n"
                    )
                elif isinstance(value, list):
                    for cache_local in value:
                        cache_archive_file = self.step_runner.build_runner.get_cache_archive_file(
                            cache_local,
                        )
                        caches[cache_archive_file] = key
                        container_meta_logger.write(
                            f"Copying local cache [{cache_archive_file}] -> docker path [{key}]\n"
                        )
                else:
                    print(f"Warning: Type {type(value)} is not supported. "
                          f"Not able to use cache functionality for {key}: {value}")
                    continue

        # add any capabilities when the container runs
        if 'cap_add' in self.config:
            if not isinstance(self.config['cap_add'], list):
                self.config['cap_add'] = [self.config['cap_add']]
            container_args['cap_add'] = self.config['cap_add']

        # allow privileged containers (used sparingly)
        if 'privileged' in self.config:
            container_args['privileged'] = self.config.get('privileged', False)

        # only expose ports for a run step if the flag is set
        if self.step_runner.build_runner.publish_ports and 'ports' in self.config:
            container_args['ports'] = self.config['ports']

        exit_code = None
        try:
            # create and start runner, linking any service containers
            image_config = DockerRunner.ImageConfig(
                _run_image,
                pull_image=pull_image,
                platform=self.config.get('platform', None)
            )
            self.runner = DockerRunner(
                image_config,
                log=self.step_runner.log,
            )
            # Figure out if we should be running systemd.  Has to happen after docker pull
            container_args["systemd"] = self.is_systemd(self.config, _run_image, self.step_runner.log)

            container_id = self.runner.start(
                links=self._service_links,
                **container_args
            )

            self.runner.restore_caches(caches)

            self.step_runner.log.write(
                f'Started build container {container_id:.10}\n'
            )

            if _cmds:
                # run each cmd
                for _cmd in _cmds:
                    container_meta_logger.write(
                        f"cmd> {_cmd}\n"
                    )
                    exit_code = self.runner.run(
                        _cmd,
                        console=container_logger,
                        # log=self.step_runner.log,
                    )
                    container_meta_logger.write(
                        f'Command "{_cmd}" exited with code {exit_code}\n'
                    )

                    if exit_code != 0:
                        break
            else:
                self.runner.attach_until_finished(container_logger)
                exit_code = self.runner.exit_code
                container_meta_logger.write(
                    f'Container exited with code {exit_code}\n'
                )

            self.runner.save_caches(caches)

        finally:
            if self.runner:
                self.runner.stop()
            if container_logger:
                container_logger.cleanup()
            if container_meta_logger:
                container_meta_logger.cleanup()

        # gather artifacts to results dir even if run has bad exit code
        if 'artifacts' in self.config:
            self._retrieve_artifacts(container_logger)

        # if we have an unsuccessful exit code abort
        if exit_code != 0:
            raise BuildRunnerProcessingError(
                "Error running build container"
            )

        context['run_runner'] = self.runner

        if 'post-build' in self.config:
            self._run_post_build(context)

    def _run_post_build(self, context):
        """
        Commit the run image and perform a docker build, prepending the run
        image to the Dockerfile.
        """
        self.step_runner.log.write(
            'Running post-build processing\n'
        )
        config = self.config['post-build']
        # post build always uses the image hash--can't pull
        if not is_dict(config):
            config = {'path': config}
        config['pull'] = False
        build_image_task = BuildBuildStepRunnerTask(
            self.step_runner,
            config,
            image_to_prepend_to_dockerfile=self.runner.commit(
                self.step_runner.log,
            )
        )
        _build_context = {}
        build_image_task.run(_build_context)
        context['run-image'] = _build_context.get('image', None)

    def cleanup(self, context):  # pylint: disable=unused-argument
        if self.runner:
            if self.runner.container:
                self.step_runner.log.write(
                    f"Destroying build container {self.runner.container['Id']:.10}\n"
                )
            self.runner.cleanup()

        if self._service_runners:
            for (_sname, _srun) in reversed(list(self._service_runners.items())):
                self.step_runner.log.write(
                    f'Destroying service container "{_sname}\"\n'
                )
                _srun.cleanup()

        if self._dockerdaemonproxy:
            self._dockerdaemonproxy.stop()

        if self._sshagent:
            self._sshagent.stop()

        if self._source_container:
            self.step_runner.log.write(
                f'Destroying source container {self._source_container:.10}\n'
            )
            self._docker_client.remove_container(
                self._source_container,
                force=True,
                v=True,
            )

    def is_systemd(self, config, image, logger):
        '''Check if an image runs systemd'''
        # Unused argument
        _ = logger

        rval = False
        if 'systemd' in config:
            rval = config.get('systemd', False)
        else:
            labels = self._docker_client.inspect_image(image).get('Config', {}).get('Labels', {})
            if labels and 'BUILDRUNNER_SYSTEMD' in labels:
                rval = labels.get('BUILDRUNNER_SYSTEMD', False)

        # Labels will be set as the string value.  Make sure we handle '0' and 'False'
        return bool(rval and rval != '0' and rval != 'False')

# Local Variables:
# fill-column: 100
# End:



"""
Copyright (C) 2020 Adobe
"""

from __future__ import print_function

import importlib.machinery
import os
import subprocess
import sys
import types

from setuptools import setup, find_packages


BASE_VERSION = '2.0'

SOURCE_DIR = os.path.dirname(
    os.path.abspath(__file__)
)
BUILDRUNNER_DIR = os.path.join(SOURCE_DIR, 'buildrunner')
VERSION_FILE = os.path.join(BUILDRUNNER_DIR, 'version.py')

THIS_DIR = os.path.dirname(__file__)


def read_requirements(filename):
    """
    :param filename:
    """
    requires = []
    dep_links = []
    try:
        with open(os.path.join(THIS_DIR, filename)) as robj:
            lnr = 0
            for line in robj.readlines():
                lnr += 1
                _line = line.strip()
                if not _line:
                    continue
                if _line.startswith('#'):
                    continue

                if _line.startswith('--extra-index-url'):
                    args = _line.split(None, 1)
                    if len(args) != 2:
                        print(
                            'ERROR: option "--extra-index-url" must have a URL argument: {}:{}'.format(
                                filename,
                                lnr
                            ),
                            file=sys.stderr,
                        )
                        continue
                    dep_links.append(args[1])

                elif _line[0].isalpha():
                    requires.append(_line)

                else:
                    print(
                        'ERROR: {}:{}:"{}" does not appear to be a requirement'.format(
                            filename,
                            lnr,
                            _line
                        ),
                        file=sys.stderr,
                    )

    except IOError as err:
        sys.stderr.write('Failure reading "{0}": {1}\n'.format(filename, err))
        sys.exit(err.errno)

    return requires, dep_links


REQUIRES, DEP_LINKS = read_requirements('requirements.txt')
requirements, dependency_links = read_requirements('test_requirements.txt')
TEST_REQUIRES = requirements
DEP_LINKS.extend(dependency_links)


def get_version():
    """
    Call out to the git command line to get the current commit "number".
    """
    if os.path.exists(VERSION_FILE):
        # Read version from file
        loader = importlib.machinery.SourceFileLoader('buildrunner_version', VERSION_FILE)
        version_mod = types.ModuleType(loader.name)
        loader.exec_module(version_mod)
        existing_version = version_mod.__version__  # pylint: disable=no-member
        print(f'Using existing buildrunner version: {existing_version}')
        return existing_version

    # Generate the version from the base version and the git commit number, then store it in the file
    try:
        cmd = subprocess.Popen(
            args=[
                'git',
                'rev-list',
                '--count',
                'HEAD',
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            encoding='utf8',
        )
        stdout = cmd.communicate()[0]
        output = stdout.strip()
        if cmd.returncode == 0:
            new_version = '{0}.{1}'.format(BASE_VERSION, output)
            print(f'Setting version to {new_version}')

            # write the version file
            if os.path.exists(BUILDRUNNER_DIR):
                with open(VERSION_FILE, 'w', encoding='utf8') as fobj:
                    fobj.write("__version__ = '%s'\n" % new_version)
            return new_version
    except Exception as exc:
        print(f'Could not generate version from git commits: {exc}')
    # If all else fails, use development version
    return f'{BASE_VERSION}.DEVELOPMENT'


with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as fobj:
    long_description = fobj.read().strip()

setup(
    name='buildrunner',
    version=get_version(),
    author='Adobe',
    author_email="noreply@adobe.com",
    license="MIT",
    url="https://github.com/adobe/buildrunner",
    description="Docker-based build tool",
    long_description=long_description,
    long_description_content_type='text/x-rst',
    packages=find_packages(exclude=['*.tests', '*.tests.*', 'tests.*', 'tests']),
    scripts=[
        'bin/buildrunner',
    ],
    package_data={
        'buildrunner': ['SourceDockerfile'],
        'buildrunner.sshagent': [
            'SSHAgentProxyImage/Dockerfile',
            'SSHAgentProxyImage/run.sh',
            'SSHAgentProxyImage/login.sh',
        ],
    },
    install_requires=REQUIRES,
    tests_require=TEST_REQUIRES,
    dependency_links=DEP_LINKS,
    test_suite='tests',
)

# Local Variables:
# fill-column: 100
# End:
